# ~/nmcog/spinnaker/associate/nealassoc/makeInheritanceHier.py
#
# Documentation by Lungsi 24 March 2020
#
# Code based on Jan2020 Chris Huyck et al.
# http://www.cwa.mdx.ac.uk/NEAL/code/assocMemJan2020.tar.gz
#
# Create an neural topology that has an inheritance hierarchy.
# In PyNN/python the topology can be accessed via this class.
#
# Invoke by creating the class, then calling createNeuralInheritanceHierarcy 
# with an inheritance hierarchy specified in python (readInheritanceClass is one
# way).

import sys
import numpy as np
#import cPickle as pickle
import pickle # for Python3
# added for nmcog
import spynnaker8 as sim
from nmcog.spinnaker.specialfunction.neal import NealCoverFunctions
from nmcog.spinnaker.specialfunction.neal import FSAHelperFunctions

class NeuralInheritanceClass:
    """
    
    +-------------------+--------+----------------------------------------------+
    | Constants         | values | comments                                     |
    +===================+========+==============================================+
    | ``neuronsPerCA``  | 10     | -                                            |
    +-------------------+--------+----------------------------------------------+
    | ``intraCAWeight`` | 0.0085 | - 0.01 is hot (every 4 ms)                   |
    |                   |        | - 0.005 does not go                          |
    +-------------------+--------+----------------------------------------------+
    | ``hierWeight``    | 0.0012 | - replaced with 1/2 fsa connection           |
    +-------------------+--------+----------------------------------------------+
    | ``primeWeight``   | 0.017  | - 0.016-0.015 threee levels of hier in 75 ms |
    |                   |        | - 0.02 too big                               |
    +-------------------+--------+----------------------------------------------+
    
    
    
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | Methods                                      |  Argument                                   | Caller                                       |
    +==============================================+=============================================+==============================================+
    | :py:meth:`.createNeuralInheritanceHierarchy` | - instance of :ref:`InheritanceReaderClass` | :ref:`NeuralThreeAssocClass`                 |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | :py:meth:`.createNeurons`                    | - number of neurons                         | :py:meth:`.createNeuralInheritanceHierarchy` |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | :py:meth:`.setRecord`                        | -                                           | :py:meth:`.createNeuralInheritanceHierarchy` |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | :py:meth:`.makeCAs`                          | -                                           | :py:meth:`.createNeuralInheritanceHierarchy` |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | :py:meth:`.makeHiersFromHier`                | - instance of :ref:`InheritanceReaderClass` | :py:meth:`.createNeuralInheritanceHierarchy` |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    | :py:meth:`.createTestAllUnits`               | - initial (first) start time                | :ref:`NeuralThreeAssocClass`                 |
    +----------------------------------------------+---------------------------------------------+----------------------------------------------+
    
    
    
    **Note:**
    
    * Specifically the instance of :ref:`InheritanceReaderClass` is the base data for :ref:`NEAL3Way`. An example base data is
    
    ::
    
        bases = {"units": ["animal", "mammal", "bird", "canary"],
                 "is-a": [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]}
        basedata = InheritanceReaderClass()
        basedata.numberUnits = len(bases["units"])
        basedata.units = bases["units"]
        basedata.isARelationships = bases["relations"]
    
    
    """
    #constants
    neuronsPerCA = 10
    intraCAWeight = 0.0085 # .01 is hot (every 4 ms) .005 doesn't go #replaced with fsa state weights
    hierWeight = 0.0012 #replaced with half fsa connections undone 
    primeWeight = 0.017#0.016-0.015 three levels of hier in 75 ms#.02 too big

    #class variables
    numCAs = -1
    cells = None
    
    #def __init__(self, simName,sim,neal,spinnVersion,fsa):
    def __init__(self, simName="spinnaker",spinnVersion=8): # modified for nmcog
        self.simName = simName
        self.spinnVersion = spinnVersion
        self.sim = sim
        self.neal = NealCoverFunctions()
        self.fsa = FSAHelperFunctions()

    def createNeurons(self,numNeurons):
        """Creates a population of `IF_cond_exp <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ neurons with :ref:`FSAHelperFunctions` ``.CELL_PARAMS``."""
        self.cells = self.sim.Population(numNeurons,self.sim.IF_cond_exp, 
                                    self.fsa.CELL_PARAMS)

    def setRecord(self):
        """Records spikes of all the neurons in the population created via :py:meth:`.createNeurons`."""
        self.cells.record(['spikes'])

    #Make a binary CA for each unit
    def makeCAs(self):
        """Makes ``numCAs`` assemblies of cell, i.e. connect neurons in the population created via :py:meth:`.createNeurons`."""
        for CA in range (0,self.numCAs):
            self.fsa.makeCA(self.cells,CA)

    #make a hierarchical relationship for each isA pair passed in
    def makeHiersFromHier(self,pythonHier):
        """Makes a hierarchical relationship for each "isA" pair (see above for a ``basedata`` example).
        For each "isA" relationship pair cell assemblies of the first element excites those of the second (using :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState`` with constant :ref:`FSAHelperFunctions` ``.HALF_ON_WEIGHT``).
        
        * Refer to :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState`` for how the cell assemblies are connected.
        
        """
        isAPairs = pythonHier.isARelationships
        numberPairs = len(isAPairs)
        for pairNumber in range (0,numberPairs):
            subCatName = isAPairs[pairNumber][0]
            superCatName = isAPairs[pairNumber][1]
            subCatNumber = pythonHier.getUnitNumber(subCatName)
            superCatNumber = pythonHier.getUnitNumber(superCatName)
            #print subCatName,superCatName,subCatNumber,superCatNumber
            #make one hierarchical relations as a half CA connection
            self.fsa.stateHalfTurnsOnState(self.cells,subCatNumber,
                                           self.cells,superCatNumber)

    #--Main way to create the heirarchy topology
    def createNeuralInheritanceHierarchy(self,inheritanceStructure):
        """Given an instance of :ref:`InheritanceReaderClass` a heirarchy topology is created.
        
        Consider the example for :ref:`NEAL3Way` whose base data is the instance such that,
        
        ::
        
            bases = {"units": ["animal", "mammal", "bird", "canary"],
                    "relations": [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]}
            basedata = InheritanceReaderClass()
            basedata.numberUnits = len(bases["units"])
            basedata.units = bases["units"]
            basedata.isARelationships = bases["relations"]
        
        Then,
        
        * A cell assembly is created for each association unit, i.e. respective assemblies for "animal", "mammal", "bird", and "canary".
        
            - A cell assembly by default has ten `neuron populations <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_. See :ref:`FSAHelperFunctions` ``.makeCA`` for more detail.
            - For this example four cell assemblies will be created.
        
        * Size of each population in an assembly comprises of some factor of the total number of association units considered. For instance, some factor times the total number of association units. The factor is named ``neuronsPerCA``.
        
            - By default, this is 10 times the total number of units.
            - For our example every population in each cell assembly will contain 40 `neurons. <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        
        * Connect (excitatory) cell assemblies among association units that form a "isA" relationship pair.
        
        Therefore, the hierarchy topology is connection among cell assemblies that form a "isA" relationship pair.
        
        **What is the practical meaning of the cell assemblies of a "hierarchy topology" in terms of extracting/visualizing the neuron units in a population?**
        
        * For our example of four base units ``["animal", "mammal", "bird", "canary"]``, i.e. there will be four cell assemblies.
        * By default ``neuronsPerCA`` = 10 (Notice that this value is equal to :ref:`FSAHelperFunctions` ``.CA_SIZE`` = 10)
        * Invoking :py:meth:`createNeurons` creates all the neuron units that will make the cell assemblies.
        
            - :py:meth:`createNeurons` creates all the neuron units in one attribute ``self.cells``
            
        * After simulating for a particular runtime, spikes from all the neurons (i.e. all neuron units in all the neuronal populations in all the assemblies) by
        
        ::
        
            allspikes = self.cells.get_data( variables=["spikes"] )
        
        
            - It should be noted that the above ``allspikes`` object is the `Neo Block <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.Block>`_
            - It is a Neo Block with only one `Segment <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.Segment>`_
        
        * Therefore,
        
            - Spike trains for all the neuron units in the cell assembly for the base unit "animal" would be.
            
                * ``allspikes.segments[0].spiketrains[0]``
                * That is, spike trains from all the ten neuron units in this cell assembly.
            
            - Similarly, for "mammal", "bird", and "canary" are respectively
            
                * ``allspikes.segments[0].spiketrains[0]``
                * ``allspikes.segments[0].spiketrains[1]``
                * ``allspikes.segments[0].spiketrains[2]``
        
        """
        self.numCAs = inheritanceStructure.numberUnits 
        numberNeurons = self.numCAs * self.neuronsPerCA
        self.createNeurons(numberNeurons)
        self.setRecord()
        self.makeCAs()
        self.makeHiersFromHier(inheritanceStructure)

    #---Test code.
    #Set up spike generators to start each unit, then stop them.
    #When run, each unit should persist, but only that unit 
    #should.
    def makeGenerator(self,genTime):
        """Set up spike generator to start each unit. ``genTime`` is a float representing the start time."""
        genTimes = genTime
        genTimeArray = {'spike_times': [genTimes]}
        spikeGen=self.sim.Population(1,self.sim.SpikeSourceArray,genTimeArray)
        return spikeGen

    #now stop all of the units after each individual unit is tested.
    def createStopAll(self,stopTimes):
        """Stop all the spikes generated, **each** generated using :py:meth:`.makeGenerator`."""
        stopTimeArray = {'spike_times': [stopTimes]}
        stopSpikeGen=self.sim.Population(1,self.sim.SpikeSourceArray,
                                         stopTimeArray)
        for unit in range (0,self.numCAs): #numCAs is numUnits
            self.fsa.turnOffStateFromSpikeSource(stopSpikeGen,self.cells,unit)

    def createTestAllUnits(self,firstTestStart):
        """The state for every cell assembly in the hierarchy topology is turned on.
        The method returns the time after turning on the state for the last cell assembly.
        """
        oneTestDuration = 100.0
        stopTimes = []
        for unit in range (0,self.numCAs): #numCAs is numUnits
            startTime = 25.0 + (unit*oneTestDuration)+firstTestStart
            generator = self.makeGenerator([startTime])
            self.fsa.turnOnStateFromSpikeSource(generator,self.cells,unit)
            lastTime =(((unit+1)*oneTestDuration)+firstTestStart)
            stopTimes = stopTimes + [lastTime]

        self.createStopAll(stopTimes)
        return lastTime

    #Create a spikeGen that primes all of the units numTests times.
    #Connect it to all of the units.
    def createTestPrimeAllUnits(self,firstTestStart,numTests):
        """Legacy."""
        oneTestDuration = 100.0
        timeBetweenSteps = 5.0
        numberPrimeSteps = 10
        for primeEpoch in range (0,self.numCAs): #numCAs is numUnits
            startTime = 25.0 + (primeEpoch*oneTestDuration)+firstTestStart
            primeTimes = []
            for primeStep in range (0,numberPrimeSteps): 
                primeTimes = primeTimes + [startTime + 
                                           (timeBetweenSteps*primeStep)]
            generator = self.makeGenerator(primeTimes)

            #self.connectGeneratorToPrimeUnit(generator,unit)
            for unit in range (0,self.numCAs): #numCAs is numUnits
                self.fsa.stimulateStateFromSpikeSource(generator,self.cells,
                                                       unit,self.primeWeight)
        return firstTestStart + (self.numCAs*oneTestDuration)

    #Create spikeGens and synapses to connect all the units
    #to be tested, and to spread up the hierarchy.
    def createTestAllInheritanceUnits(self,firstTestStart):
        """Legacy."""
        testFullTime = self.createTestPrimeAllUnits(firstTestStart,self.numCAs)
        #this assumes that both of the functions have the same testduration.
        otherTestFullTime = self.createTestAllUnits(firstTestStart)
        print("bob", testFullTime, otherTestFullTime)
        return otherTestFullTime

    #This is a test to see if some neurons are firing.
    def createSimpleTest(self):
        """Legacy."""
        primeTimes = [5]
        primeArray = {'spike_times': [primeTimes]}
        generator=self.sim.Population(1,self.sim.SpikeSourceArray,primeArray)

        self.fsa.turnOnStateFromSpikeSource(generator,self.cells,0)

    #def printInheritanceHier(self,fileName):
    #    """Legacy."""
    #    self.cells.write_data(fileName) #self.cells.printSpikes(fileName) depreciated

    




# ~/nmcog/spinnaker/specialfunction/neal/timerClass.py
#
# Documentation by Lungsi 6 April 2020
#
# This makes several different types of timers and provides mechanisms
# for the timers to effect other states and be turned on by other 
# states, spike sources and neurons.
#

# added for nmcog
import spynnaker8 as sim
from .nealCoverClass import NealCoverFunctions
from .stateMachineClass import FSAHelperFunctions

#---- Code
class TimerClass:
    """This is the class for constructing different types of timers. It also provide functions for the timers
    
    * to effect other states
    * to be turned on by other states, spike sources and neurons.
    
    """
    forwardWeight = 0.004
    oneNeuronFullWeight = 0.08
    oneNeuronHalfWeight = 0.05#0.1#0.04

    #def __init__(self, simName,sim,neal,spinnVersion,fsa,numberStates):
    def __init__(self, simName="spinnaker", spinnVersion=8):
        self.simName = simName
        self.sim = sim
        self.neal = NealCoverFunctions()
        self.spinnVersion = spinnVersion
        self.fsa = FSAHelperFunctions()
        #self.numberStates = numberStates # commented out for nmcog
        #self.timerCells = self.createNeurons(numberStates) # commented out for nmcog

    #create the neurons for the words and the states.
    def createNeurons(self,numberStates):
        """For a given number of states (under consideration) the number of timer cells equals
        the number of states times the size of the cell assembly (given by :ref:`FSAHelperFunctions` ``.CA_SIZE``).
        
        By default :ref:`FSAHelperFunctions` ``.CA_SIZE`` = 10. Therefore, if the number of possible states is two
        then there will be twenty timer cells.
        
        **Note:**
        
        * A timer cell is `PyNN's IF_cond_exp <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_
        * The timer cells are collectively the `PyNN's population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        
        """
        # added for nmcog
        self.numberStates = numberStates # self.timerCells are no longer made at instantiation
        #
        numberTimerCells = numberStates * self.fsa.CA_SIZE
        self.timerCells = self.sim.Population(numberTimerCells,self.sim.IF_cond_exp,
                                              self.fsa.CELL_PARAMS)
        self.setupRecording()
        
        return self.timerCells

    #-----start timer functions
    def connectSpikeGenToStart(self,spikeGen):
        """See :ref:`FSAHelperFunctions` ``.turnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(spikeGen,self.timerCells,0) # timercells are components of first (i.e. 0) cell assembly

    def stateStartsTimer(self,stateNeurons,state):
        """See :ref:`FSAHelperFunctions` ``.stateTurnsOnState``"""
        self.fsa.stateTurnsOnState(stateNeurons,state,self.timerCells,0) # timercells are components of first (i.e. 0) cell assembly

    #undone not in test
    def oneNeuronStartsTimer(self,fromNeurons,fromNeuron):
        """See :ref:`FSAHelperFunctions` ``.oneNeuronStimulatesState``
        This differs from :py:meth:`.oneNeuronHalfStartsTimer` because here ``self.oneNeuronFullWeight`` is used.
        """
        self.fsa.oneNeuronStimulatesState(fromNeurons,fromNeuron,
            self.timerCells,0,self.oneNeuronFullWeight)

    def oneNeuronHalfStartsTimer(self,fromNeurons,fromNeuron):
        """See :ref:`FSAHelperFunctions` ``.oneNeuronStimulatesState``
        This differs from :py:meth:`.oneNeuronStartsTimer` because here ``self.oneNeuronHalfWeight`` is used.
        """
        self.fsa.oneNeuronStimulatesState(fromNeurons,fromNeuron,
            self.timerCells,0,self.oneNeuronHalfWeight)

    def stateHalfStartsTimer(self,stateNeurons,state):
        """See :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState``"""
        self.fsa.stateHalfTurnsOnState(stateNeurons,state,self.timerCells,0) # timercells are components of first (i.e. 0) cell assembly

    #----stop functions
    #undone needs a test and move to timer
    def neuronStopsTimer(self,stopNeurons,stopNeuron):
        """See :ref:`FSAHelperFunctions` ``.oneNeuronTurnsOffState``"""
        for timerCA in range (0,self.numberStates):
            self.fsa.oneNeuronTurnsOffState(stopNeurons,stopNeuron,
                                        self.timerCells,timerCA)

    #undone needs a test
    def stateStopsTimer(self,stateNeurons,state):
        """See :ref:`FSAHelperFunctions` ``.stateTurnsOffState``"""
        for timerCA in range (0,self.numberStates):
            self.fsa.stateTurnsOffState(stateNeurons,state,
                                        self.timerCells,timerCA)

    #---timer output functions
    #The last state of a timer starts a state
    def timerStartsState(self,stateNeurons,state):
        """See :ref:`FSAHelperFunctions` ``.stateStimulatesState``"""
        #stateTurnsOnState can take a couple of steps.
        #self.fsa.stateTurnsOnState(self.timerCells,self.numberStates-1,
        #                           stateNeurons,state)
        self.fsa.stateStimulatesState(self.timerCells,self.numberStates-1,
                                      stateNeurons,state,0.02)

    def timerPreventsState(self,stateNeurons,state):
        """See :ref:`FSAHelperFunctions` ``.stateTurnsOffState``"""
        for CA in range (0,self.numberStates-1):
            self.fsa.stateTurnsOffState(self.timerCells,CA,
                                   stateNeurons,state)

    #undone needs a test
    def timerActivatesState(self,stateNeurons,state,wt):
        """See :ref:`FSAHelperFunctions` ``.stateStimulatesState``"""
        for CA in range (0,self.numberStates-1):
            self.fsa.stateStimulatesState(self.timerCells,CA,stateNeurons,
                                          state,wt)


    #---create timer functions
    #The timer goes from state to state and the last stops on its own
    def makeStopTimerSynapses(self):
        """See :ref:`FSAHelperFunctions` ``.stateStimulatesState`` and :ref:`FSAHelperFunctions` ``.stateTurnsOffState``"""
        for CA in range (0,self.numberStates-1):
            self.fsa.makeCA(self.timerCells,CA)
            self.fsa.stateStimulatesState(self.timerCells,CA,self.timerCells,
                                             CA+1,self.forwardWeight)
            self.fsa.stateTurnsOffState(self.timerCells,CA+1,self.timerCells,CA)
            
    #The timer goes from state to state and the last stops on its own
    #Only one state is on at at time
    def makeOneStateStopTimerSynapses(self):
        """See :ref:`FSAHelperFunctions` ``.makeCA``, :ref:`FSAHelperFunctions` ``.stateStimulatesState``,
        and :ref:`FSAHelperFunctions` ``.stateTurnsOffState``"""
        for CA in range (0,self.numberStates-1):
            self.fsa.makeCA(self.timerCells,CA)
            self.fsa.stateStimulatesState(self.timerCells,CA,self.timerCells,
                                             CA+1,self.forwardWeight)
            self.fsa.stateTurnsOffState(self.timerCells,CA+1,self.timerCells,CA)
        for toCA in range (0,self.numberStates-1):
            for fromCA in range (toCA+2,self.numberStates):
                self.fsa.stateTurnsOffState(self.timerCells,fromCA,
                                            self.timerCells,toCA)

    #The timer goes from state to state and the last stops on its own
    #If the state that turns it on remains on, this timer will not restart
    #until its done
    def makeStopNoRestartTimerSynapses(self):
        """
        Once a timercell is created using :py:meth:`.createNeurons` returning a `neuron population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        composed of :math:`n \\times 10` cells where *n* is its number of possible states.
        
        This function connects all its possible states.
        
        Consider a timercell, *timerX* with five states then        
        
        ::
    
                    sssssssssssssssssssssssssssss
                    s ccccccccc     s ccccccccc s     ccccccccc       ccccccccc
                    s c xxxxx c     s c xxxxx c s     c xxxxx c       c xxxxx c
                    V V x   V c     V V x   V c s     V x   V c       V x   V c
            ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo
            o timerX CA o   o timerX CA o   o timerX CA o   o timerX CA o   o timerX CA o
            o  state-0  o   o  state-1  o   o  state-2  o   o  state-3  o   o  state-4  o
            ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo
                    ^               ^               ^        s
                    ssssssssssssssssssssssssssssssssssssssssss
        
        
        * all the states are numbers 0, 1, ..., 4 (= 5-1)
        * each state is represented by a cell assembly using :ref:`FSAHelperFunctions` ``.makeCA``
        
            - here four cell assemblies one for each of the five states
            
        * starting from the first state, 0 (above, state-0),
        
            - each state (i.e. its cell assembly) stimulates (xxxxxx) its immediate succeeding state using :ref:`FSAHelperFunctions` ``.stateStimulatesState``
            
                - notice that the first state is not simulated
                
            - each state (i.e. its cell assembly) is turned-off (cccccc) by its immediate succeeding state using :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
            
                - notice that the last state is not turned-off using :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        
        * starting from the third state, 2 (above, state-2)
        
            - each state turns off (ssssss) all the preceeding states
            
                - notice that the last state does not turn-off its preceeding states.
        
        
        **Note:**
        
        * 10 in :math:`n \\times 10` is the default :ref:`FSAHelperFunctions` ``.CA_SIZE`` = 10
        * The timer cells are collectively the `PyNN's population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        * A timer cell is `PyNN's IF_cond_exp <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_
        
        """
        for CA in range (0,self.numberStates-1):
            self.fsa.makeCA(self.timerCells,CA)
            self.fsa.stateStimulatesState(self.timerCells,CA,self.timerCells,
                                             CA+1,self.forwardWeight)
            self.fsa.stateTurnsOffState(self.timerCells,CA+1,self.timerCells,CA)

        #have the later CAs prevent all the earlier ones
        for fromCA in range (2,self.numberStates-1):
            for toCA in range (0,fromCA-1):
                self.fsa.stateTurnsOffState(self.timerCells,fromCA,
                                            self.timerCells,toCA)

        
    def setupRecording(self):
        """Records spikes and voltages (i.e. analog signals)."""
        self.timerCells.record({'spikes','v'})

    def printResults(self,fileName):
        """Legacy."""
        fileName = fileName + ".pkl"
        self.timerCells.write_data(fileName)
        
#-- Access Functions
    def getTimerCells(self):
        """Returns ``self.timercells"""
        return self.timerCells






# ~/nmcog/spinnaker/associate/nealassoc/readUnitFile.py
#
# Code based on Jan2020 Chris Huyck et al.
# http://www.cwa.mdx.ac.uk/NEAL/code/assocMemJan2020.tar.gz
#
# Read in from a file a list of units terminated by @@@@@.
# Thease are stored into the units array.
#
# Unlike implementation by Chris Huyck et al.
# nmcog does not use this class to read .txt file the contents of which
# are transformed into a structured data.
#
# Although, the structued data can be made in nmcog without this class
# because attributes defined here are inherited by classes calling it
# this class has been incorporated in nmcog.

class UnitReaderClass:
    """
    **Note:**
    
    * Unlike implementation by Chris Huyck et al. nmcog does not use this class to read .txt file
    * The structured data are created within the working class.
    * However, the reading ability of :py:class:`UnitReaderClass` has not been disabled therefore it can still be used like Chris Huyck et al.
    * Although, in the current implementation of any class within nmcog creates its structured data the reason :py:class:`UnitReaderClass` is because of some original scripts (Chris Huyck et al.) within nmcog that invokes methods (like the :py:meth:`getUnitNumber`) within this class whose internal has instantiated attributes.
    
    """
    #instance variables
    units = []
    numberUnits = -1

    #--Functions for reading in the units
    #read lines until you get one starting with an @
    #each of these is a primitive unit in the hierarchy
    def readUnits(self,handle):
        """Read lines until you get one starting with an "@" Each of these is a primitive unit in the hierarchy."""
        unitListDone = False
        while (not unitListDone):
            line = handle.readline()
            if (line[0]=='@'):
                unitListDone = True
            else:
                unitName = line.strip() #take off the \n
                self.units = self.units + [unitName]
    
    def inUnits(self,checkUnit):
        """Check if a unit in question is a unit."""
        done = False
        unitListOffset = 0
        while (not done):
            if (checkUnit == self.units[unitListOffset]):
                return True
            unitListOffset = unitListOffset + 1
            if (unitListOffset == self.numberUnits):
                done = True
        return False

    def getUnitNumber(self,checkUnit):
        """Return the total number of units."""
        for resultUnit in range (0,self.numberUnits):
            if (checkUnit == self.units[resultUnit]):
                return resultUnit
        print ("error ", checkUnit , " not in unit array")


    #---top level functions
    def createViaRead(self,fileName):
        """Top level function."""
        fileHandle = open(fileName, 'r')
        self.readUnits(fileHandle)
        self.numberUnits = len(self.units)
        fileHandle.close()

    def readUnitFile(self,fileName):
        """Reads a text file."""
        inputFileName = fileName+".txt"
        self.createViaRead(inputFileName)




# ~/nmcog/spinnaker/specialfunction/neal/nealCoverClass.py
#
# Documentation by Lungsi 25 March 2020
#
import spynnaker8 as sim # added for nmcog

class NealCoverFunctions:
    """Functions to isolate differences between different list synapse constructors.
    
    +----------------------------------+--------------------------------------------------------------------------------------------------------------------------------+------------------------------------+
    | Methods                          |  Argument                                                                                                                      | Caller                             |
    +==================================+================================================================================================================================+====================================+
    | :py:meth:`.nealProjection`       | - projection from a `PyNN based spike source <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ | usually to connect between two     |
    |                                  |or a `PyNN based Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_                                  |populations between cell assemblies |
    |                                  | - projection into another `PyNN based Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_            |                                    |
    |                                  | - a list of connection parameter                                                                                               |                                    |
    |                                  | - string describing the connection type; "excitatory" or "inhibitory"                                                          |                                    |
    +----------------------------------+--------------------------------------------------------------------------------------------------------------------------------+------------------------------------+
    | :py:meth:`.sameProjectionType`   | - two projections                                                                                                              | :py:meth:`.nealApplyProjections`   |
    +----------------------------------+--------------------------------------------------------------------------------------------------------------------------------+------------------------------------+
    | :py:meth:`.nealApplyProjections` | - (class must be instantiated because ``self.projections`` would have been the argument if this method had one)                |                                    |
    +----------------------------------+--------------------------------------------------------------------------------------------------------------------------------+------------------------------------+
    
    
    **Note:**
    
    * The `PyNN based Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ is usually a population within a cell assembly.
    * The list of connection parameter is of the form such that
    
        - first element is the projection source (i.e. `PyNN based spike source <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ or `PyNN based Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_)
        - second element is the projection target (usually another `PyNN based Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_)
        - third element is a float type value representing weight value
        - fourth element is the ``self.DELAY`` (default = 0.1)
    
    * The two projections must each be a list of the form ``[preNeurons, postNeurons, inhExc]`` or ``[preNeurons, postNeurons, inhExc, connectorList]``.
    
    """

    projections = []
    
    #def __init__(self, simName,sim,spinnVersion):
    def __init__(self, simName="spinnaker", spinnVersion=8): # modified for nmcog
        self.DELAY = 1.0
        self.simulator = simName
        self.sim = sim
        self.spinnVersion = spinnVersion

    def nealProjection(self,preNeurons,postNeurons,connectorList,inhExc):
        """Given ``preNeurons``, ``postNeurons``, ``connectorList``, and ``inhExc`` (string, "inhibitory" or "excitatory") these four arguments are put
        into a list and reordered such that ``inhExc`` goes in front of ``connectorList``.
        The list is then appended to the class attribute ``self.projections``, a list.
        """
        newProjection = [preNeurons,postNeurons,inhExc,connectorList]
        self.projections.append(newProjection)

    def sameProjectionType(self,projectionA,projectionB):
        """Checks if two given projections (between two cell assemblies) are of the same type. Returns Boolean. Note these projections are **not** ``self.projections``."""
        if ((projectionA[0] == projectionB[0]) and # pre Neurons 
            (projectionA[1] == projectionB[1]) and # post Neurons
            (projectionA[2] == projectionB[2])):   # inh or exc
            return True
        return False

    #collect the projections into projections that are pre post and type
    #specific.  Write those out in one fromList
    def nealApplyProjections(self):
        """Collect the projections (i.e. ``self.projections``) into projections that are pre post and type specific (via :py:meth:`.sameProjectionType`).
        And write them into one fromList."""
        while (len(self.projections) > 0):
            projectionNumber = 0
            #get the first projection and collect all projections like it
            firstProjection = self.projections[0]
            connList = firstProjection[3]
            self.projections.remove(firstProjection)
            while (projectionNumber < len(self.projections)):
                if (self.sameProjectionType(self.projections[projectionNumber],
                           firstProjection)):
                    connList = connList + self.projections[projectionNumber][3]
                    #print projectionNumber,len(connList)
                    self.projections.remove(self.projections[projectionNumber])
                else:
                    projectionNumber = projectionNumber + 1
                    
            #actually put out the projections
            fromListConnector = self.sim.FromListConnector(connList)
            preNeurons = firstProjection[0]
            postNeurons = firstProjection[1]
            inhExc = firstProjection[2]
            if ((self.simulator=="spinnaker") and (self.spinnVersion==7)):
                self.sim.Projection(preNeurons, postNeurons, fromListConnector,
                                   target=inhExc)
            elif ((self.simulator=="nest") or
              ((self.simulator=="spinnaker") and (self.spinnVersion==8))):
                    self.sim.Projection(preNeurons, postNeurons, fromListConnector,
                                   receptor_type=inhExc)
            
            else: print("bad simulator nealProjection")


# ~/nmcog/spinnaker/specialfunction/neal/stateMachineClass.py
#
# Documentation by Lungsi 25 March 2020
#
#This is the class for construcing State Machines.  The individual CAs 
#(both inputs and states) run at 5 ms.  See the notes for running faster or 
#slower. 
#There are test functions at the end. 
#Note that to externally activate a CA State, you should only send
#excitatory connections to the first 8 neurons.
#Using the nealNRPCover so no nealParams
#
import numpy as np
import pickle

from .nealCoverClass import NealCoverFunctions

# added for nmcog
import spynnaker8 as sim

class FSAHelperFunctions:
    """This is the class for constructing State Machines.  The individual CAs  (both inputs and states) run at 5 ms. 
    
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | Methods                                     |  Parameters Used                                                                                   |
    +=============================================+====================================================================================================+
    | :py:meth:`.turnOnStateFromSpikeSource`      | - ``CA_SIZE``, ``CA_INHIBS``, ``INPUT_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``               |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.halfTurnOnStateFromSpikeSource`  | - ``CA_SIZE``, ``CA_INHIBS``, ``HALF_INPUT_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``          |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.turnOffStateFromSpikeSource`     | - ``CA_SIZE``, ``CA_INHIBS``, ``ONE_NEURON_STOPS_CA_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY`` |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stimulateStateFromSpikeSource`   | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.inhibitStateFromSpikeSource`     | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.turnOnOneNeuronFromSpikeSource`  | - ``INPUT_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``                                           |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.turnOffOneNeuronFromSpikeSource` | - ``CA_STOPS_CA_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``                                     |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.inhibitOneNeuronFromSpikeSource` | - :ref:`NealCoverFunctions` ``.DELAY``                                                             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronTurnsOnState`           | - ``CA_SIZE``, ``CA_INHIBS``, ``INPUT_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``               |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronHalfTurnsOnState`       | - ``CA_SIZE``, ``CA_INHIBS``, ``ONE_HALF_ON_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``         |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronTurnsOffState`          | - ``CA_SIZE``, ``CA_INHIBS``, ``ONE_NEURON_STOPS_CA_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY`` |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronHalfTurnsOffState`      | - ``CA_SIZE``, ``CA_INHIBS``, ``ONE_HALF_ON_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``         |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronStimulatesState`        | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronStimulatesOneNeuron`    | - :ref:`NealCoverFunctions` ``.DELAY``                                                             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronTurnsOnOneNeuron`       | - ``INPUT_WEIGHT``                                                                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronInhibitsOneNeuron`      | - :ref:`NealCoverFunctions` ``.DELAY``                                                             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronHalfTurnsOnOneNeuron`   | - ``ONE_HALF_ON_ONE_WEIGHT``                                                                       |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.oneNeuronInhibitsState`          | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateTurnsOnOneNeuron`           | - ``CA_SIZE``, ``CA_INHIBS``, ``STATE_TO_ONE_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``        |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateTurnsOnOneRBSNeuron`        | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateHalfTurnsOnOneNueron`       | - ``CA_SIZE``, ``CA_INHIBS``, ``HALF_ON_ONE_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``         |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateStimulatesOneNeuron`        | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateInhibitsOneNeuron`          | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateTurnsOnState`               | - ``CA_SIZE``, ``CA_INHIBS``, ``FULL_ON_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateTurnsOnStateSlow`           | - ``CA_SIZE``, ``CA_INHIBS``, ``FULL_ON_WEIGHT_SLOW``, :ref:`NealCoverFunctions` ``.DELAY``        |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateTurnsOffState`              | - ``CA_SIZE``, ``CA_INHIBS``, ``CA_STOPS_CA_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``         |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateHalfTurnsOnState`           | - ``CA_SIZE``, ``CA_INHIBS``, ``HALF_ON_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateHalfTurnsOffState`          | - ``CA_SIZE``, ``CA_INHIBS``, ``HALF_ON_WEIGHT``, :ref:`NealCoverFunctions` ``.DELAY``             |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateStimulatesState`            | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.stateInhibitsState`              | - ``CA_SIZE``, :ref:`NealCoverFunctions` ``.DELAY``                                                |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    | :py:meth:`.getCAConnectors`                 | - ``CA_SIZE``, ``CA_INHIBS``, :ref:`NealCoverFunctions` ``.DELAY``                                 |
    |                                             | - ``INTRA_CA_WEIGHT``, ``INTRA_CA_TO_INHIB_WEIGHT``, ``INTRA_CA_FROM_INHIB_WEIGHT``                |
    +---------------------------------------------+----------------------------------------------------------------------------------------------------+
    
    
    """
    
    #def __init__(self, simName, sim, neal,spinnVersion = None):
    def __init__(self, simName="spinnaker", spinnVersion = 8): # modified for nmcog
        self.simName = simName
        self.sim = sim
        self.neal = NealCoverFunctions()
        if spinnVersion is None:
            self.spinnVersion = -1
        else:
            self.spinnVersion = spinnVersion
        self.initParams()
        

    #FSA Parmaeters
    def initParams(self):
        """
    
        +--------------------------------+------------+--------------------------------------------------------------------+
        | Constants                      | values     | comments                                                           |
        +================================+============+====================================================================+
        | ``CA_SIZE``                    | 10         | - This will (almost certainly) not work with a different sized CA. |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``CA_INHIBS``                  | 2          | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``INPUT_WEIGHT``               | 0.12       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``HALF_INPUT_WEIGHT``          | 0.008      | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``INTRA_CA_TO_INHIB_WEIGHT``   | 0.002      | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``INTRA_CA_FROM_INHIB_WEIGHT`` | 0.15       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``CA_STOPS_CA_WEIGHT``         | 0.15       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``ONE_NEURON_STOPS_CA_WEIGHT`` | 1.0        | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``ONE_NEURON_HALF_CA_WEIGHT``  | 0.02       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``INTRA_CA_WEIGHT``            | 0.025      | - for `sPyNNaker <https://spinnakermanchester.github.io/>`_        |
        |                                | 0.022      | - for `NEST <https://www.nest-simulator.org/>`_                    |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``FULL_ON_WEIGHT``             | 0.01       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``FULL_ON_WEIGHT_SLOW``        | 0.0022     | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``HALF_ON_WEIGHT``             | 0.0012     | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``HALF_ON_ONE_WEIGHT``         | 0.002      | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``STATE_TO_ONE_WEIGHT``        | 0.01       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``ONE_HALF_ON_WEIGHT``         | 0.016      | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``ONE_HALF_ON_ONE_WEIGHT``     | 0.08       | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        | ``CELL_PARAMS``                | dictionary | -                                                                  |
        |                                |            |                                                                    |
        | 'v_thresh'                     | -48.0      | -                                                                  |
        |                                |            |                                                                    |
        | 'v_reset'                      | -70.0      | -                                                                  |
        |                                |            |                                                                    |
        | 'tau_refrac'                   | 2.0        | -                                                                  |
        |                                |            |                                                                    |
        | 'tau_syn_E'                    | 5.0        | -                                                                  |
        |                                |            |                                                                    |
        | 'tau_syn_I'                    | 5.0        | -                                                                  |
        |                                |            |                                                                    |
        | 'v_rest'                       | -65.0      | -                                                                  |
        |                                |            |                                                                    |
        | 'i_offset'                     | 0.0        | -                                                                  |
        +--------------------------------+------------+--------------------------------------------------------------------+
        
        """
        self.CA_SIZE = 10  #This will (almost certainly) not work with a 
                     #different sized CA.
        self.CA_INHIBS = 2

        #normal weights
        self.INPUT_WEIGHT = 0.12
        self.HALF_INPUT_WEIGHT = 0.08
        self.INTRA_CA_TO_INHIB_WEIGHT = 0.002
        #if you over inhib on my spinnaker, it fires.
        self.INTRA_CA_FROM_INHIB_WEIGHT = 0.15 
        self.CA_STOPS_CA_WEIGHT = 0.15
        self.ONE_NEURON_STOPS_CA_WEIGHT = 1.0 

        self.ONE_NEURON_HALF_CA_WEIGHT = 0.02

        if (self.simName =="spinnaker"):
            self.INTRA_CA_WEIGHT = 0.025
        elif (self.simName == "nest"):
            self.INTRA_CA_WEIGHT = 0.022 
        else:
            self.INTRA_CA_WEIGHT = 0.022 

        self.FULL_ON_WEIGHT = 0.01 
        self.FULL_ON_WEIGHT_SLOW = 0.0022
        self.HALF_ON_WEIGHT = 0.0012
        self.HALF_ON_ONE_WEIGHT = 0.002
        #undone rbs self.STATE_TO_ONE_WEIGHT = .015
        self.STATE_TO_ONE_WEIGHT = .01

        self.ONE_HALF_ON_WEIGHT = 0.016
        self.ONE_HALF_ON_ONE_WEIGHT = 0.08
        #self.ONE_NEURON_STARTS_CA_WEIGHT = self.INPUT_WEIGHT


        self.CELL_PARAMS = {'v_thresh':-48.0, 'v_reset' : -70.0, 
                                'tau_refrac': 2.0 , 'tau_syn_E': 5.0,  
                                'tau_syn_I' : 5.0, 
                                'v_rest' : -65.0,'i_offset':0.0}

    #--------Finite State Automata Functions ------------
    #states can be turned and off by a spikeSource, and can be stimulate or
                   #inhibited by one.
    #states can be turned on and off by a neuron, and can stimulate or inhibit
                    #one
    #states can stimulate or inhibit neurons

    #states can be turned on and off by a state, and can stimulate or inhibit
                    #them.  States can also slow turn on other states, and
                    #half turn them on


    #---Functions that turn on states by one item
    #-- Function to ignite a state from a spike source
    #-- Uses INPUT_WEIGHT
    #use this when your using a spikesource
    def turnOnStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **ON** the state of the desired cell assembly.
        ``toNeurons`` specifies the type of `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within the cell assembly.
        
        +-----------------+-------------+---------------+------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``INPUT_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,self.INPUT_WEIGHT,
                                      self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'excitatory')

    def halfTurnOnStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **1/2 ON** the state of the desired cell assembly.
        ``toNeurons`` specifies the type of `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within the cell assembly.
        
        +-----------------+-------------+---------------+-----------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``HALF_INPUT_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+-----------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,self.HALF_INPUT_WEIGHT,
                                      self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'excitatory')

    def turnOffStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **OFF** the state of the desired cell assembly.
        ``toNeurons`` specifies the type of `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within the cell assembly.
        
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``ONE_NEURON_STOPS_CA_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,
                        self.ONE_NEURON_STOPS_CA_WEIGHT,self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'inhibitory')

    def stimulateStateFromSpikeSource(self,spikeSource, toNeurons, toCA, weight):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method stimulates the state of the desired cell assembly with the given weight value.
        ``toNeurons`` specifies the type of `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within the cell assembly.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,weight,
                                      self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'excitatory')

    def inhibitStateFromSpikeSource(self,spikeSource, toNeurons, toCA, weight):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method inhibits the state of the desired cell assembly.
        ``toNeurons`` specifies the type of `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within the cell assembly.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,weight,
                                      self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'inhibitory')

    def turnOnOneNeuronFromSpikeSource(self,spikeSource, toNeurons, toNeuron):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **ON** the state of a desired `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        within a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_.
        
        +-----------------+------------------+--------------------------------------+
        | Parameters used | ``INPUT_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+------------------+--------------------------------------+
        
        """
        connector = []
        connector = connector + [(0,toNeuron,self.INPUT_WEIGHT,self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'excitatory')

    #undone needs a test
    def turnOffOneNeuronFromSpikeSource(self,spikeSource, toNeurons, toNeuron):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **OFF** the state of a desired `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        within a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_.
        
        +-----------------+------------------------+--------------------------------------+
        | Parameters used | ``CA_STOPS_CA_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+------------------------+--------------------------------------+
        
        """
        connector = []
        connector = connector + [(0,toNeuron,self.CA_STOPS_CA_WEIGHT,self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'inhibitory')

    #undone
    def inhibitOneNeuronFromSpikeSource(self,spikeSource, toNeurons, toNeuron,
                                        weight):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method inhibits the state of a desired `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        within a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value.
        
        +-----------------+--------------------------------------+
        | Parameters used | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+--------------------------------------+
        
        """
        connector = []
        connector = connector + [(0,toNeuron,weight,self.neal.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'inhibitory')

    #---states can be turned on and off by a neuron, can stimulate or 
    #inhibit one, and can half turn on one.
    def oneNeuronTurnsOnState(self,fromNeurons,fromNeuron, toNeurons, toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **ON** the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly.
        
        +-----------------+-------------+---------------+------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``INPUT_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,self.INPUT_WEIGHT,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'excitatory')

    def oneNeuronHalfTurnsOnState(self,fromNeurons,fromNeuron, toNeurons, toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **1/2 ON** the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly.
        
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``ONE_HALF_ON_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,self.ONE_HALF_ON_WEIGHT,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'excitatory')

    def oneNeuronTurnsOffState(self,fromNeurons, fromNeuron, toNeurons, toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **OFF** the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly.
        
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``ONE_NEURON_STOPS_CA_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,
                                      self.ONE_NEURON_STOPS_CA_WEIGHT,
                                      self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')

    def oneNeuronHalfTurnsOffState(self,fromNeurons,fromNeuron, toNeurons, toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **1/2 OFF** the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly.
        
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``ONE_HALF_ON_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,self.ONE_HALF_ON_WEIGHT,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'inhibitory')

    def oneNeuronStimulatesState(self,fromNeurons,fromNeuron, toNeurons, toCA, 
                                weight):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method stimulates the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly
        with the given weight value.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,weight,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'excitatory')


    #the rbs has these, but they're not accurate weights for what they
    #say they do.  So, leave them out for now.
    #def oneNeuronHalfTurnsOnState(self,fromNeurons,fromNeuron,toNeurons,toCA):
    #def oneNeuronHalfTurnsOnOneNeuron(self,fromNeurons,fromCA,toNeurons,toCA):
    #def stateHalfTurnsOnOneNeuron(self,fromNeurons,fromCA,toNeurons,toNeuron):

    #Neurons can also directly interact with each other.
    def oneNeuronStimulatesOneNeuron(self,fromNeurons,fromNeuron, toNeurons, 
                                     toNeuron,weight):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method stimulates the state
        of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_ of a
        desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value.
        
        +-----------------+--------------------------------------+
        | Parameters used | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+--------------------------------------+
        
        """
        connector = []
        connector = connector + [(fromNeuron,toNeuron,weight,self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'excitatory')

    def oneNeuronTurnsOnOneNeuron(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **ON** the state
        of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_ of a
        `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a desired cell assembly.
        
        +-----------------+------------------+
        | Parameters used | ``INPUT_WEIGHT`` |
        +-----------------+------------------+
        
        """
        self.oneNeuronStimulatesOneNeuron(fromNeurons,fromCA,toNeurons,toCA, 
                                       self.INPUT_WEIGHT)

    def oneNeuronInhibitsOneNeuron(self,fromNeurons,fromNeuron, 
                                   toNeurons,toNeuron, weight):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method inhibits the state
        of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_ of a
        desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value.
        
        +-----------------+--------------------------------------+
        | Parameters used | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+--------------------------------------+
        
        """
        connector = []
        connector = connector + [(fromNeuron,toNeuron,weight,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'inhibitory')

    def oneNeuronHalfTurnsOnOneNeuron(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method turns **1/2 ON** the state
        of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_ of a
        `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a desired cell assembly.
        
        +-----------------+----------------------------+
        | Parameters used | ``ONE_HALF_ON_ONE_WEIGHT`` |
        +-----------------+----------------------------+
        
        """
        self.oneNeuronStimulatesOneNeuron(fromNeurons,fromCA,toNeurons,toCA, 
                                       self.ONE_HALF_ON_ONE_WEIGHT)

    def oneNeuronInhibitsState(self,fromNeurons,fromNeuron, toNeurons, toCA, 
                               weight):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, this method inhibits the state
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a desired cell assembly
        with the given weight value.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        """
        connector = []
        for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,weight,
                                      self.neal.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'inhibitory')


    #states can stimulate or inhibit neurons
    def stateTurnsOnOneNeuron(self,fromNeurons,fromCA,toNeurons,toNeuron):
        """From a given a `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **ON** the state of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_.
        
        +-----------------+-------------+---------------+-------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``STATE_TO_ONE_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+-------------------------+--------------------------------------+
        
        Connection between the cell assemblies is as shown
        
        ::
        
                    fromCA
            ooooooooooooooooooooooo
            o                     o
            o                     o                         toNeuron
            o (0) (2) (4) (6) (8) o                          ( x )
            o      +              o  STATE_TO_ONE_WEIGHT       ^
            o      +++++++++++++++o+++++++++++++++++++++++++++++
            o                     o
            o (1) (3) (5) (7) (9) o
            o                     o
            o                     o
            ooooooooooooooooooooooo
        
        
        **Note:**
        
        * A cell assembly is composed of CA_SIZE number of `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ such that each unit is of same kind.
        * Therefore, the arguments ``fromNeurons`` and ``toNeurons`` (plural) refers to the kind of neuron population in the from- and to-assemblies.
        * That is, ``fromNeurons`` and ``toNeurons`` do not indicate some specific neuron population within respective assemblies.
        * The specificiation of a unit (from and to) a cell assembly is done by the connector list of tuples.
        * This specification is such that the ``fromNeuron`` and ``toNeuron`` are only in reference to units with excitatory connections.
        
            - Referring to illustration in :py:meth:`.getCAConnectors` then in the above illustration ``fromNeuron`` and ``toNeuron`` (singular) will be <<0>> to <<7>>
            - <<8>> and <<9>> are not the source for the connection.
            - In theory ``toNeuron`` can be in the same cell assembly as those of ``fromNeuron``.
        
        """
        connector = []
        #uses STATE_TO_ONE_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            connector=connector+[(fromNeuron,toNeuron,
                                  self.STATE_TO_ONE_WEIGHT,
                                  self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    #states can stimulate or inhibit neurons
    def stateTurnsOnOneRBSNeuron(self,fromNeurons,fromCA,toNeurons,toNeuron):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **ON** the state of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value 0.015.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateTurnsOnOneNeuron`.
        
        """
        connector = []
        #uses STATE_TO_ONE_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            connector=connector+[(fromNeuron,toNeuron,
                                  0.015,
                                  self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    def stateHalfTurnsOnOneNueron(self,fromNeurons,fromCA,toNeurons,toNeuron):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **1/2 ON** the state of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_.
        
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``HALF_ON_ONE_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateTurnsOnOneNeuron`.
        
        """
        connector = []
        #uses STATE_TO_ONE_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            connector=connector+[(fromNeuron,toNeuron,
                                  self.HALF_ON_ONE_WEIGHT,
                                  self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    def stateStimulatesOneNeuron(self,fromNeurons,fromCA,toNeurons,toNeuron,
                                 weight):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method stimulates the state of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateTurnsOnOneNeuron`.
        
        """
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            connector=connector+[(fromNeuron,toNeuron,weight,
                                  self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')


    def stateInhibitsOneNeuron(self,fromNeurons,fromCA,toNeurons,toNeuron, 
                              weight):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method inhibits the state of another `neuron unit <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#standard-cell-types>`_
        of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ with the given weight value.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateTurnsOnOneNeuron` with the exception that this is "inhibitory".
        
        """
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            connector=connector+[(fromNeuron,toNeuron,weight,
                                  self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')


    #states can be turned on and off by a state, and can stimulate or inhibit
                    #them.  States can also slow turn on other states, and
                    #half turn them on

    #Function to turn on one state from another
    #Call with fromPopulation, fromCA, toPopulation and toCA
    def stateTurnsOnState(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **ON** the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly.
        
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``FULL_ON_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateHalfTurnsOnState`.
        
        """
        connector = []
        #uses FULL_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.FULL_ON_WEIGHT,
                                          self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    #When stateTurnsOnState and the preState remains on, the post
    #state runs hot.
    def stateTurnsOnStateSlow(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method slowly turns **ON** the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly.
        
        +-----------------+-------------+---------------+-------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``FULL_ON_WEIGHT_SLOW`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+-------------------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateHalfTurnsOnState`.
        
        """
        connector = []
        #uses FULL_ON_WEIGHT_SLOW
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.FULL_ON_WEIGHT_SLOW,
                                          self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')


    #---- One State or other set of neurons turns off another
    #-- Uses CA_STOPS_CA_WEIGHT
    def stateTurnsOffState(self,fromNeurons, fromCA, toNeurons, toCA):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **OFF** the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly.
        
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``CA_STOPS_CA_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+------------------------+--------------------------------------+
        
        Connection between the cell assemblies is as shown
        
        ::
        
                    fromCA                                        toCA
            ooooooooooooooooooooooo                      ooooooooooooooooooooooo
            o                     o                      o                     o
            o                     o                      o                     o
            o (0) (2) (4) (6) (8) o                      o (0) (2) (4) (6) (8) o
            o      -              o  CA_STOPS_CA_WEIGHT  o  ^   ^   ^   ^   ^  o
            o      ---------------o----------------------o-------------------  o
            o                     o                      o  v   v   v   v   v  o
            o (1) (3) (5) (7) (9) o                      o (1) (3) (5) (7) (9) o
            o                     o                      o                     o
            o                     o                      o                     o
            ooooooooooooooooooooooo                      ooooooooooooooooooooooo
        
        
        **Note:**
        
        * A cell assembly is composed of CA_SIZE number of `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ such that each unit is of same kind.
        * Therefore, the arguments ``fromNeurons`` and ``toNeurons`` (plural) refers to the kind of neuron population in the from- and to-assemblies.
        * That is, ``fromNeurons`` and ``toNeurons`` do not indicate some specific neuron population within respective assemblies.
        * The specificiation of a unit (from and to) a cell assembly is done by the connector list of tuples.
        * This specification is such that the ``fromNeuron`` and ``toNeuron`` are from every unit within a cell assembly.
        
            - Thus, unlike the illustration in :py:meth:`.stateHalfTurnsOnState` in the above illustration ``fromNeuron`` and ``toNeuron`` (singular) will be <<0>> to <<9>>
            - See the illustration in :py:meth:`.getCAConnectors` for referring to <<0>> to <<9>> 
        
        
        """
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,
                                              self.CA_STOPS_CA_WEIGHT,
                                              self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')

    #--Functions when two inputs are needed to turn on a third
    #-- Two states are needed to turn on a third.  
    #-- This connects one of the inputs to the the third.
    #-- Uses HALFs_ON_WEIGHT
    def stateHalfTurnsOnState(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **1/2 ON** the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly.
        
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``HALF_ON_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        
        Connection between the cell assemblies is as shown
        
        ::
        
                    fromCA                                    toCA
            ooooooooooooooooooooooo                  ooooooooooooooooooooooo
            o                     o                  o                     o
            o                     o                  o                     o
            o (0) (2) (4) (6) (8) o                  o (0) (2) (4) (6) (8) o
            o      +              o  HALF_ON_WEIGHT  o  ^   ^   ^   ^      o
            o      +++++++++++++++o++++++++++++++++++o+++++++++++++++      o
            o                     o                  o  v   v   v   v      o
            o (1) (3) (5) (7) (9) o                  o (1) (3) (5) (7) (9) o
            o                     o                  o                     o
            o                     o                  o                     o
            ooooooooooooooooooooooo                  ooooooooooooooooooooooo
        
        
        **Note:**
        
        * A cell assembly is composed of CA_SIZE number of `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ such that each unit is of same kind.
        * Therefore, the arguments ``fromNeurons`` and ``toNeurons`` (plural) refers to the kind of neuron population in the from- and to-assemblies.
        * That is, ``fromNeurons`` and ``toNeurons`` do not indicate some specific neuron population within respective assemblies.
        * The specification of a unit (from and to) a cell assembly is done by the connector list of tuples.
        * This specification is such that the ``fromNeuron`` and ``toNeuron`` are only in reference to units with excitatory connections.
        
            - Referring to illustration in :py:meth:`.getCAConnectors` then in the above illustration ``fromNeuron`` and ``toNeuron`` (singular) will be <<0>> to <<7>>
            - <<8>> and <<9>> are neither the source nor the target for the connection.
        
        """
        connector = []
        #uses HALF_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.HALF_ON_WEIGHT,
                                          self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    def stateHalfTurnsOffState(self,fromNeurons,fromCA,toNeurons,toCA):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method turns **1/2 OFF** the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly.
        
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``HALF_ON_WEIGHT`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateHalfTurnsOnState`.
        
        """
        connector = []
        #uses HALF_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.HALF_ON_WEIGHT,
                                          self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')

    def stateStimulatesState(self,fromNeurons,fromCA,toNeurons,toCA,weight):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method stimulates the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly with the given weight value.
        
        +-----------------+-------------+---------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+---------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateHalfTurnsOnState`.
        
        """
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                    weight, self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    def stateInhibitsState(self,fromNeurons, fromCA, toNeurons, toCA, wt):
        """From a given `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ in a cell assembly,
        this method inhibits the state of a desired `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        within a desired cell assembly with the given weight value.
        
        +-----------------+-------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | :ref:`NealCoverFunctions` ``.DELAY`` |
        +-----------------+-------------+--------------------------------------+
        
        **Note:**
        
        * For illustration on how the connections are made see :py:meth:`.stateTurnsOffState` with the exception that this is "inhibitory".
        
        """
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,
                                              wt,self.neal.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')



    #---Create a CA that will persistently fire.
    #-- Assumes neurons in the same population
    #-- Uses INTRA_CA_WEIGHT
    def getCAConnectors(self, CA):
        """Create excitatory connector list and inhibitory connector list for a given cell assembly.
        The method returns a list such that the first element is the excitatory connector list.
        
        Regardless of excitatory or inhibitory connector lists, the most basic element of a connector list is a
        `PyNN Connector <http://neuralensemble.org/docs/PyNN/reference/connectors.html#pyNN.connectors.FromListConnector>`_ tuple
        of the form *(pre_idx, post_idx, weight, delay)* where *pre_idx* is the index
        (i.e. order in the `Population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_, not the ID) of the presynaptic neuron,
        *post_idx* is the index of the postsynaptic neuron, and *weight*, *delay* are the synaptic parameters.
        
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        | Parameters used | ``CA_SIZE`` | ``CA_INHIBS`` | ``INTRA_CA_WEIGHT``            | :ref:`NealCoverFunctions` ``.DELAY`` |
        |                 |             |               | ``INTRA_CA_TO_INHIB_WEIGHT``   |                                      |
        |                 |             |               | ``INTRA_CA_FROM_INHIB_WEIGHT`` |                                      |
        +-----------------+-------------+---------------+--------------------------------+--------------------------------------+
        
        By default a cell assembly size ``CA_SIZE`` = 10, which is the number of `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ in each assembly.
        
        These ten neuron populations are such that
        
        * eight of them project excitatory connections, lets call them <<0>> to <<7>>
        * two of them project inhibitory connections, lets call them <<8>> and <<9>>
        
        Three kinds of connections are made,
        
        1. Excitatory connection from one neuron unit to another unit that projects excitatory connection (but not itself). Below shows projections from <<0>>.
        
        ::
        
            Connection Weight is INTRA_CA_WEIGHT.
            
             .-.      .-.      .-.      .-.      .-.
            ( 0 )    ( 2 )    ( 4 )    ( 6 )    ( 8 )
             '-'      '-'      '-'      '-'      '-'
              +        ^        ^        ^
              + ++++++++++++++++++++++++++
              v        v        v        v
             .-.      .-.      .-.      .-.      .-.
            ( 1 )    ( 3 )    ( 5 )    ( 7 )    ( 9 )
             '-'      '-'      '-'      '-'      '-'
        
        2. Excitatory connection from neuron unit to another that project inhibitory connection. Below shows projections from <<0>>.
        
        ::
        
            Connection Weight is INTRA_CA_TO_INHIB_WEIGHT.
            
             .-.      .-.      .-.      .-.      .-.
            ( 0 )    ( 2 )    ( 4 )    ( 6 )    ( 8 )
             '-'      '-'      '-'      '-'      '-'
              +                                   ^
              +++++++++++++++++++++++++++++++++++++
                                                  v
             .-.      .-.      .-.      .-.      .-.
            ( 1 )    ( 3 )    ( 5 )    ( 7 )    ( 9 )
             '-'      '-'      '-'      '-'      '-'
        
        3. Inhibitory connection from neuron unit to other unit that project excitatory connection (but not to units that project inhibitory connection including itself). Below shows projections from <<8>>.
        
        ::
        
            Connection Weight is INTRA_CA_FROM_INHIB_WEIGHT.
            
             .-.      .-.      .-.      .-.      .-.
            ( 0 )    ( 2 )    ( 4 )    ( 6 )    ( 8 )
             '-'      '-'      '-'      '-'      '-'
              ^        ^        ^        ^        -
              -------------------------------------
              v        v        v        v
             .-.      .-.      .-.      .-.      .-.
            ( 1 )    ( 3 )    ( 5 )    ( 7 )    ( 9 )
             '-'      '-'      '-'      '-'      '-'
        
        **Note:**
        
        * The two kinds of excitatory connections are collected in one list, the excitatory connector list.
        
        """
        connector = []
        #excitatory turn each other on
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (CA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (CA*self.CA_SIZE)
                if (toNeuron != fromNeuron):
                    connector = connector + [(fromNeuron,toNeuron,
                        self.INTRA_CA_WEIGHT, self.neal.DELAY)]

        #excitatory turn on inhibitory
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            fromNeuron = fromOffset + (CA*self.CA_SIZE)
            for toOffset in range (self.CA_SIZE-self.CA_INHIBS,self.CA_SIZE):
                toNeuron = toOffset + (CA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,
                        self.INTRA_CA_TO_INHIB_WEIGHT, self.neal.DELAY)]


        #inhibitory slows excitatory 
        inhConnector = []
        for fromOffset in range (self.CA_SIZE-self.CA_INHIBS,self.CA_SIZE):
            fromNeuron = fromOffset + (CA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (CA*self.CA_SIZE)
                inhConnector = inhConnector + [(fromNeuron,toNeuron,
                        self.INTRA_CA_FROM_INHIB_WEIGHT, self.neal.DELAY)]

        return[connector,inhConnector]

    def makeCA(self,neurons, CA):
        """Given a `neuron population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ and cell assembly parameters
        this method gets excitatory and inhibitory connectors (from :py:meth:`.getCAConnectors`) for the presumptive cell assembly.
        
        * Because the default a cell assembly size ``CA_SIZE`` = 10, there are ten `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_
        
        ::
        
            oooooooooooooooooooooooooooooooooooooooooooooooo
            o                                              o
            o   .-.      .-.      .-.      .-.      .-.    o
            o  ( 0 )    ( 2 )    ( 4 )    ( 6 )    ( 8 )   o
            o   '-'      '-'      '-'      '-'      '-'    o
            o                                              o
            o        cell assembly of CA_SIZE = 10         o
            o                                              o
            o   .-.      .-.      .-.      .-.      .-.    o
            o  ( 1 )    ( 3 )    ( 5 )    ( 7 )    ( 9 )   o
            o   '-'      '-'      '-'      '-'      '-'    o
            o                                              o
            oooooooooooooooooooooooooooooooooooooooooooooooo
        
        
        * These neuron units are connected using :py:meth:`.getCAConnectors` put together via :ref:`NealCoverFunctions` ``.nealProjection``
        * The source and target `neuron populations <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ are one and the same.
        
        Therefore, a cell assembly here is
        
        * **one** `PyNN neuron population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ 
        * with ten `neuron units <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ within the `population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
        * and the ten neuron units are connected (see :py:meth:`.getCAConnectors` for how the connections are made).
        
        """
        connectors = self.getCAConnectors(CA)
        self.neal.nealProjection(neurons,neurons,connectors[0],'excitatory')
        self.neal.nealProjection(neurons,neurons,connectors[1],'inhibitory')

    def nopMakeCA(self,neurons, CA):
        """Legacy."""
        connector = []
        fromNeuron = CA
        toNeuron = CA + 1
        connector = connector + [(fromNeuron,toNeuron, 1.2, self.neal.DELAY)]
        self.neal.nealProjection(neurons,neurons,connector,'excitatory')
        
#------test functions
    #initialize the simulator. 
    def testInit(self):
        """Legacy."""
        #print "spin" or nest
        self.sim.setup(timestep=self.neal.DELAY,
                    min_delay=self.neal.DELAY,
                    max_delay=self.neal.DELAY, debug=0)


    def testCreateTwoInputs(self):
        """Legacy."""
        inputSpikeTimes0 = [10.0]
        inputSpikeTimes1 = [50.0]
        spikeArray0 = {'spike_times': [inputSpikeTimes0]}
        spikeGen0=self.sim.Population(1,self.sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')
        spikeArray1 = {'spike_times': [inputSpikeTimes1]}
        spikeGen1=self.sim.Population(1, self.sim.SpikeSourceArray, spikeArray1,
                                   label='inputSpikes_1')

        return [spikeGen0,spikeGen1]

    def testCreateNeurons(self):
        """Legacy."""
        if (self.simName == 'nest'):
            numNeurons = self.CA_SIZE * 3
        elif (self.simName == 'spinnaker'):
            numNeurons = 100
        
        cells=self.sim.Population(numNeurons,self.sim.IF_cond_exp,self.CELL_PARAMS)

        return cells

    def testSetupRecording(self,cells):
        """Legacy."""
        if  ((self.neal.simulator == 'nest') or 
             ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 8))):
            cells.record({'spikes','v'})
        elif ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 7)):
            cells.record()

    def test3StateFSA(self, firstSpikeGenerator, secondSpikeGenerator,
                      stateCells):
        """Legacy."""
        #Build the FSA
        self.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
        self.turnOnStateFromSpikeSource(secondSpikeGenerator,stateCells,1)
        self.makeCA(stateCells,0)
        self.makeCA(stateCells,1)
        self.makeCA(stateCells,2)
        self.stateHalfTurnsOnState(stateCells,0,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,0)
        #comment below out to check state 0 alone does not turn on state 2
        self.stateHalfTurnsOnState(stateCells,1,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,1)

    def testRunFSA(self,duration):
        """Legacy."""
        self.sim.run(duration)

    def testPrintPklSpikes(self,fileName):
        """Legacy."""
        fileHandle = open(fileName)
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    print( neuronNum, spikes[spike] )
        fileHandle.close()
    

    def testPrintResults(self,simCells):
        """Legacy."""
        simCells.write_data('temp.pkl',['spikes'])


# =============================================================================
# ~/spinnaker/cognitivemap/nealplacesystem.py
#
# created January 2020 Lungsi
#
# =============================================================================
import spynnaker8 as sim

# for plotting
import quantities as pq
import matplotlib.pyplot as plt
#import matplotlib.patches as mpatches
#from matplotlib import gridspec
#from matplotlib import cm

from .nealmentalmap.placecellsysClass import PlaceCellSystemClass
from nmcog.spinnaker.specialfunction.neal import NealCoverFunctions

class NEALPlaceSystem(object):
    """
    Consider nobject = 2 and nplaces = 3
    
    +------------------+---------------------+------------------+-----------------+
    | Example use case | objectsTOplaces     | findPlaceFor     | findObjectFor   |
    +==================+=====================+==================+=================+
    | 1 (default)      | None                | None             | None            |
    +------------------+---------------------+------------------+-----------------+
    | 2                | [(0,1),(1,0),(1,2)] | None             | None            |
    +------------------+---------------------+------------------+-----------------+
    | 3                | [(0,1),(1,0),(1,2)] | 1 (for object 1) | None            |
    +------------------+---------------------+------------------+-----------------+
    | 4                | [(0,1),(1,0),(1,2)] | None             | 2 (for place 2) |
    +------------------+---------------------+------------------+-----------------+
    
    """
    def __init__(self, nobjects=None, nplaces=None, objectsTOplaces=None, find="all"):
        self.nobjects = nobjects
        self.nplaces = nplaces
        self.objectsTOplaces = objectsTOplaces
        #
        self.neal = NealCoverFunctions()
        #
        self.inputTimes = self.__generateSpikeTimes()
        if find=="all":
            self.questions, self.answers = self.__run_all( findkey="for-object" )
            qes, ans = self.__run_all( findkey="for-place" )
            self.questions.update( qes )
            self.answers.update( ans )
        else:
            for findkey, findval in find.items():
                self.questions, self.answers = self.__run( findkey, findval )
    
    def __run_all(self, findkey=None):
        """Given a key, "for-object" or "for-place" this function runs
        for all its respective elements and returns a dictionary."""
        if findkey=="for-object":
            allfinds = self.nobjects
        elif findkey=="for-place":
            allfinds = self.nplaces
        questions = {}
        answers = {}
        for findFor in range(allfinds):
            q_dict, a_dict = self.__run( findkey, findFor )
            if findFor==0: # create initial dictionary for respective findkey
                questions.update( q_dict )
                answers.update( a_dict )
            else: # append on previously created dictionary for respective findkey
                questions[findkey].update( q_dict[findkey] )
                answers[findkey].update( a_dict[findkey] )
        return questions, answers
    
    def __run(self, findkey, findval):
        # one: {findkey: val}
        # all: findkey and findval is i of the loop
        sim.setup( timestep = self.neal.DELAY,
                   min_delay = self.neal.DELAY,
                   max_delay = self.neal.DELAY, debug=0)
        self.spikeSource = self.__makeSpikeSource_from_inputTimes()
        self.__createCogmap()
        self.cogmap.sourceStartsAutomaton( self.spikeSource[0] )
        #
        self.__bindObjectsToPlaces( {findkey: findval} )
        if findkey=="for-object":
            self.__retrievePlaceForObject( findval )
        elif findkey=="for-place":
            self.__retrieveObjectForPlace( findval )
        #
        self.neal.nealApplyProjections()
        sim.run( self.inputTimes[-1]+500 )
        #
        q_dict, a_dict = self.__getdata( findkey, findval )
        #
        #self.cogmap.printCogMapNets()
        #sim.end()
        #
        return q_dict, a_dict
        
    
    def __createCogmap(self):
        """Creates the cognitive map for the place cell system using :ref:`PlaceCellSystemClass`."""
        self.cogmap = PlaceCellSystemClass()
        self.cogmap.createAutomaton()
        self.cogmap.createObjects( self.nobjects )
        self.cogmap.connectObjects()
        self.cogmap.createPlaces( self.nplaces )
        self.cogmap.connectPlaces()
        self.cogmap.setupCogMapRecording()
        self.cogmap.makeLearningSynapses()
        self.cogmap.connectAutomaton()
        
    def __makeSpikeSource_from_inputTimes(self):
        """Returns a list of spike sources. The number of spike source determined by
        the number of ``self.inputTimes`` produced by :py:meth:`.generateSpikeTimes`."""
        spikeArrays = [ {"spike_times": [ anInputTime ]} for anInputTime in self.inputTimes ]
        spikeGens = [ sim.Population( 1, sim.SpikeSourceArray, spikeArrays[i],
                                      label="inputSpikes_"+str(i+1) )
                        for i in range( len(spikeArrays) ) ]
        return spikeGens
    
    # Private function
    def __generateSpikeTimes(self):
        """Creates array of spike times required for making spike source.
        
        * The first spike time is for starting the automaton.
        * The second for the first object with a place
        * The rest for objects with a place but **not** for objects with no place
        * The last spike time for retrieving and quering about an object.
        
        """
        inputTimes = [ 10. ] # for starting the automaton
        if self.objectsTOplaces is not None: # does not include objects with no place,
            oTp = self.__object_place_tuple_list( self.objectsTOplaces )
            for i in range( 0, len(oTp)+1 ):
                if i==0: # for first object
                    inputTimes.append( 50. )
                elif i==1: # for second object
                    inputTimes.append( inputTimes[-1] * inputTimes[0] )
                else: # rest object, last is for retrieval and query about an object
                    inputTimes.append( inputTimes[-1] + 500. )
        return inputTimes
    
    # Private function
    def __object_place_tuple_list(self, objectsTOplaces):
        """Not all the objects in the ``objectTOplaces`` are in a place.
        Return a list of tuple that only includes objects in a place."""
        oTp = []
        [oTp.append(tupl) for tupl in objectsTOplaces if len(tupl)!=1]
        return oTp
    
    # Private function
    def __rearrange_oTp_for_findval(self, find, oTp):
        """This step is needed otherwise Chris Huyk et al's function does not behave as
        advertised, i.e. after about two objects querying the succeeding objects does
        not result in answers eventhough by definition these objects are bound to a place.
        I found that for any object (or place) in question if they are moved up the list
        i.e. the will receive the spikesource[1] (the first is reserved for initiating the
        automaton) then querying any bound object will return an answer."""
        y = []; tmp = [] # first and temporary list
        for findkey, findval in find.items():
            if findkey=="for-object":
                indx = 0 # first index of a tupl
            elif findkey=="for-place":
                indx = 1 # second index of a tupl
            [y.append(tupl) if tupl[indx]==findval else tmp.append(tupl) for tupl in oTp]
        return y+tmp
    
    # Private function
    def __bindObjectsToPlaces(self, find):
        """Binds objects to place using :ref:`PlaceCellSystemClass` ``.sourceTurnsOnBind``, :ref:`PlaceCellSystemClass` ``.sourceTurnsOnPlaceOn``, and
        :ref:`PlaceCellSystemClass` ``.sourceTurnsOnObjectOn``."""
        if self.objectsTOplaces is not None:
            oTp = self.__object_place_tuple_list( self.objectsTOplaces )
            reordered_oTp = self.__rearrange_oTp_for_findval(find, oTp)
            for i in range( len(reordered_oTp) ):
                sourceIndx = i+1 # first,0, spike source is reserved for automaton
                objPlacePair = reordered_oTp[i]
                self.cogmap.sourceTurnsOnBind( self.spikeSource[sourceIndx] )
                self.cogmap.sourceTurnsOnPlaceOn( objPlacePair[1], self.spikeSource[sourceIndx] )
                self.cogmap.sourceTurnsOnObjectOn( objPlacePair[0], self.spikeSource[sourceIndx] )
    
    def __retrievePlaceForObject(self, findPlaceFor):
        """Where is the object?"""
        self.cogmap.sourceTurnOnRetrievePlaceFromObject( self.spikeSource[-1] )
        self.cogmap.sourceTurnsOnObjectQuery( self.spikeSource[-1], findPlaceFor )
            
    def __retrieveObjectForPlace(self, findObjectFor):
        """What object are in the place?"""
        self.cogmap.sourceTurnOnRetrieveObjectFromPlace( self.spikeSource[-1] )
        self.cogmap.sourceTurnsOnPlaceQuery( self.spikeSource[-1], findObjectFor )
        
    def __getdata(self, findkey, findval):
        if findkey=="for-object":
            spks_qes = self.cogmap.queryOnObjectCells.get_data( variables=["spikes"] )
            spks_ans = self.cogmap.answerPlaceCells.get_data( variables=["spikes"] )
        elif findkey=="for-place":
            spks_qes = self.cogmap.queryOnPlaceCells.get_data( variables=["spikes"] )
            spks_ans = self.cogmap.answerObjectCells.get_data( variables=["spikes"] )
        q_dict = { findkey: { str(findval): spks_qes } }
        a_dict = { findkey: { str(findval): spks_ans } }
        return q_dict, a_dict
    
    # Private function
    def __generate_range(self, n):
        r = []
        for i in range(n):
            if i==0:
                indx_start = 0
                indx_end = self.cogmap.fsa.CA_SIZE
            else:
                indx_start = indx_end
                indx_end = indx_end + self.cogmap.fsa.CA_SIZE
            r.append( [indx_start, indx_end] )
        return r
    
    def plot(self, obj=None, pla=None):
        """."""
        plt.close() # close any previous figure
        allrange_nplaces = self.__generate_range( self.nplaces )
        allrange_nobjects = self.__generate_range( self.nobjects )
        if obj is not None:
            fig, ( allsps ) = plt.subplots(self.nplaces, 2, sharex=True)
            for i in range(self.nplaces): # second column of subplots for places
                for j in range( allrange_nplaces[i][0], allrange_nplaces[i][1] ): # shift the spike train by subtracting with last inputTime
                    allsps[i][1].eventplot( self.answers["for-object"][str(obj)].segments[0].spiketrains[j] - self.inputTimes[-1]*pq.ms )
                allsps[i][1].title.set_text('Place-'+str(i))
                allsps[i][1].set_yticks( [] )
                if i==(self.nplaces-1):
                    allsps[i][1].set(xlabel="time (ms)")
            #
            for i in range(self.nobjects): # first column of subplots for objects
                for j in range( allrange_nobjects[i][0], allrange_nobjects[i][1] ): # shift the spike train by subtracting with last inputTime
                    allsps[i][0].eventplot( self.questions["for-object"][str(obj)].segments[0].spiketrains[j] - self.inputTimes[-1]*pq.ms )
                allsps[i][0].title.set_text('Object-'+str(i))
                allsps[i][0].set_yticks( [] )
                if i==(self.nobjects-1):
                    allsps[i][0].set(xlabel="time (ms)")
            plt.subplots_adjust( hspace=.5 ) # spacing for the each subplot title
            plt.show()
        elif pla is not None:
            fig, ( allsps ) = plt.subplots(self.nobjects, 2, sharex=True)
            for i in range(self.nobjects): # second column of subplots for objects
                for j in range( allrange_nobjects[i][0], allrange_nobjects[i][1] ): # shift the spike train by subtracting with last inputTime
                    allsps[i][1].eventplot( self.answers["for-place"][str(pla)].segments[0].spiketrains[j] - self.inputTimes[-1]*pq.ms )
                allsps[i][1].title.set_text('Object-'+str(i))
                allsps[i][1].set_yticks( [] )
                if i==(self.nobjects-1):
                    allsps[i][1].set(xlabel="time (ms)")
            #
            for i in range(self.nplaces): # first column of subplots for places
                for j in range( allrange_nplaces[i][0], allrange_nplaces[i][1] ): # shift the spike train by subtracting with last inputTime
                    allsps[i][0].eventplot( self.questions["for-place"][str(pla)].segments[0].spiketrains[j] - self.inputTimes[-1]*pq.ms )
                allsps[i][0].title.set_text('Place-'+str(i))
                allsps[i][0].set_yticks( [] )
                if i==(self.nplaces-1):
                    allsps[i][0].set(xlabel="time (ms)")
            plt.subplots_adjust( hspace=.5 ) # spacing for the each subplot title
            plt.show()


# =============================================================================
# ~/spinnaker/discriminate/BuonomanoMerzenich.py
#
# created November 2019 Lungsi
#
# =============================================================================

import copy
import spynnaker8 as sim
import numpy as np
import quantities as pq
# for plotting
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import gridspec
#from mpl_toolkits.axes_grid1.inset_locator import inset_axes
#import matplotlib.image as mpimg

class BuoMerz(object):
    """`SpiNNaker (sPyNNaker) <https://spinnakermanchester.github.io/>`_ implementation of Temporal Information Processing.
    
    **Architecture:**
    
    * Stimulus is sent to a network representing Douglas and Martin's `(1989) <https://doi.org/10.1162/neco.1989.1.4.480>`_ cortical IV and III layers.
    * Keeping with experimental observations the excitatory:inhibitory element is approximately 4:1.
    * The constituent excitatory and inhibitory elements are randomly connected (Buonomano &amp; Merzenich, 1997).
    * The elements are based on Integrate-and-fire current units (Buonomano &amp; Merzenich, 1995).
    
    The connection is based on Buonomano & Merzenich 1997 (Fig 3, p135)
        
    +----------------------+------------------+-----------------+------------------+-----------------+
    | Populations          | excitatory (III) | excitatory (IV) | inhibitory (III) | inhibitory (IV) |
    +======================+==================+=================+==================+=================+
    | **excitatory (III)** | 18/200 = 0.09    | -               | 12/200 = 0.06    | -               |
    +----------------------+------------------+-----------------+------------------+-----------------+
    | **excitatory (IV)**  | 18/120 = 0.15    | 15/120 = 0.125  | 12/120 = 0.1     | 10/120 = 0.083  |
    +----------------------+------------------+-----------------+------------------+-----------------+
    | **inhibitory (III)** | 8/50 = 0.16      | -               | 6/50 = 0.12      | -               |
    +----------------------+------------------+-----------------+------------------+-----------------+
    | **inhibitory (IV)**  | -                | 6/30 = 0.2      | -                | 4/30 = 0.133    |
    +----------------------+------------------+-----------------+------------------+-----------------+
    | **Input**            | -                | 15/100 = 0.15   | -                | 10/100 = 0.1    |
    +----------------------+------------------+-----------------+------------------+-----------------+
        
    **Note:**
        
    * The numerator correspond to the number of units (elements) within the recieving population.
    * The denominator is the convergence number of presynaptic inputs for each unit.
    * The decimal fraction are the probability values for the `FixedProbabilityConnector <http://neuralensemble.org/docs/PyNN/reference/connectors.html#built-in-connectors>`_ function.
        
    **Comments on connection with output layer:**
        
    The excitatory (III) population connects to the output layer `(1995) <https://doi.org/10.1126/science.7863330>`_ such that it proxies adaptation. The output layer populations recieves signal only at the end of the stimulus, i.e. during second pulse.
        
    Due to limitations with `SpNNaker8 <http://spinnakermanchester.github.io/>`_ here the connection is implemented such that it is made from the start of the simulation. However, only responses from the populations in the output layer at the end of the stimulus is returned.
    
    **References:**

    * Buonomano, D. V., &amp; Merzenich, M. M. (1995). *Temporal information transformed into a spatial code by a neural network with realistic properties*. Science: 1028-1030. DOI: `10.1126/science.7863330 <https://doi.org/10.1126/science.7863330>`_
    * Buonomano, D. V., &amp; Merzenich, M. M. (1997). Temporal Information Processing: A Computational Role for Paired-Pulse Facilitation and Slow Inhibition. In J. W. Donahoe &amp; V. Packard Dorsel (Eds.), *Neural-Networks Models of Cognition* (pp. 129-139). Netherlands, Amsterdam: Elsevier Science B. V.
    
    """
    # Network parameters (see Buonomano & Merzenich 1997, Fig 3, p 135)
    n_input = 100 # number of unit in input layer
    n_ex4 = 120   # number of excitatory units in layer-4
    n_inh4 = 30   # number of inhibitory units in layer-4
    n_ex3 = 200   # number of excitatory units in layer-3
    n_inh3 = 50   # number of inhibitory units in layer-3
    #w_ex3_out = 0 # initial weight
    #d_ex3_out = 0.1 # delay
    n_out = 100
    # Wiring of network components are done in __connect()
    # Unit (cell) parameters (see Buonomano & Merzenich 1995)
    ex_cell_parameters = {
            'v_rest':   -65.0,  # Resting membrane potential in mV.
            'cm':         1.5,  # Capacity of the membrane in 1.0 nF (default)
            'tau_m':     20.0,  # Membrane time constant in 20 ms (default)
            'tau_refrac': 0.1,  # Duration of refractory period in ms.
            'tau_syn_E':  4.0,  # Rise time of the excitatory synaptic alpha function in 0.5 ms (default)
            'tau_syn_I':  80.0,  # Rise time of the inhibitory synaptic alpha function in 0.5 ms (default)
            'i_offset':   0.0,  # Offset current in nA
            'v_reset':  -65.0,  # Reset potential after a spike in mV.
            'v_thresh': -40.0,  # Spike threshold in -50.0 mV (default)
            }
    inh_cell_parameters = copy.deepcopy(ex_cell_parameters)
    inh_cell_parameters["v_thresh"] = -50.
    # output-layer receives input from excitatory population of layer 3
    # Buonomano and Merzenich tested the network for intervals between 30 and 330 ms with 10 ms steps
    #dual_pulse_intervals = np.linspace(30, 330, num=31) # 30ms : 10ms : 330ms => 300/10 + 1 = 31 ms

    def __init__(self, intervals):
        self.dual_pulse_intervals = intervals # [80, 130, 180, 230, 280]
        self.str_dual_pulse_intervals = [ str(stimulus) for stimulus in self.dual_pulse_intervals]  # ["80", "130", "180", "230", "280"] for plotting
        self.data_for_all_intervals = {}
        for self.inter_pulse_interval in intervals:
            sim.setup(1)
            self.setup_inputchannel( [ self.inter_pulse_interval ] )
            self.input_src = self.__gen_input_src()
            [self.popIn, self.layer4, self.layer3, self.output] = self.__create_layers()
            self.__connect()
            self.__record()
            sim.run( self.runtime )
            [neo_popIn, neo_ex4, neo_inh4, neo_ex3, neo_inh3, neo_out] = self.__getdata()
            #sim.reset() # mod line
            sim.end()
            self.data_for_all_intervals.update(
                    { str(self.inter_pulse_interval): { "origin": self.stim_origin,
                                                        "end": self.stim_end,
                                                        "popIn": neo_popIn, "out": neo_out,
                                                        "ex4": neo_ex4, "inh4": neo_inh4,
                                                        "ex3": neo_ex3, "inh3": neo_inh3 }
                                                        } )
            
    def get_results(self):
        """
        Gets the recorded `Neo <https://neo.readthedocs.io/en/latest/>`_ objects.
        The only exception is for the output layer.
        Because of the reasons given above, for the output layer rather than a Neo object its `SpikeTrain <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.SpikeTrain>`_ are returned.
        They are the responses at the end of the stimulus.
        To help visualizing them along with `SpikeTrain <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.SpikeTrain>`_ from other layers a padding is also returned.
        
        Returns data as a dictionary such that
        
        +----------+---------------------------------------------------------------------------------------------+
        | Key      | Value                                                                                       |
        +==========+=============================================================================================+
        | "popIn"  | Neo object for input layer; contains ``SpikeTrain`` and ``Analogsignal``                    |
        +----------+---------------------------------------------------------------------------------------------+
        | "ex4"    | Neo object for excitatory population LayerIV; contains ``SpikeTrain`` and ``Analogsignal``  |
        +----------+---------------------------------------------------------------------------------------------+
        | "inh4"   | Neo object for inhibitory population LayerIV; contains ``SpikeTrain`` and ``Analogsignal``  |
        +----------+---------------------------------------------------------------------------------------------+
        | "ex3"    | Neo object for excitatory population LayerIII; contains ``SpikeTrain`` and ``Analogsignal`` |
        +----------+---------------------------------------------------------------------------------------------+
        | "inh3"   | Neo object for inhibitory population LayerIII; contains ``SpikeTrain`` and ``Analogsignal`` |
        +----------+---------------------------------------------------------------------------------------------+
        | "out"    | Dictionary for output layer with populations under the keys "out"+stimuli, say "out80"      |
        +----------+---------------------------------------------------------------------------------------------+
        | "origin" | float type representing start of the runtime                                                |
        +----------+---------------------------------------------------------------------------------------------+
        | "end"    | float type representing end of the runtime                                                  |
        +----------+---------------------------------------------------------------------------------------------+
        
        **Note:**
        
        Each population in the output layer (value of "out") is also a dictionary with keys "spiketrains" and "placeholder_axes".
        
        * The value for "placeholder_axes" is a list containing two lists.
        
            - The first represent t-axis from zero to start of second pulse of the stimuli.
            - The second is an array of zeros whose size is the same as the first.
        
        * The value for "spiketrains" is the `SpikeTrain` for *all* the populations 
        
        """
        return self.data_for_all_intervals


    def setup_inputchannel(self, all_intervals):
        """Given a stimulus, i.e. interval values this function creates an input channel for it.
        
        The input channel is such that
        
        ::
        
                         ___|-interval/stimulus-|___
                        |   |                   |   |
          origin________|   |___________________|   |________end
               |---t0---|-T-|                   |-T-|---t0---|
               |------------------runtime--------------------|
        
        
        Note that the above dual-pulse train represented in ``self.input_channel`` is used to create a spiking version using the private function ``__gen_input_src()``.
        
        In addition to the ``self.input_channel`` this function also sets the attributes: ``self.stim_origin``, ``self.stim_end``, and ``self.runtime``.
        
        **Note:**
        
        * Interval value or the inter-pulse interval is the temporal information carried by the stimulus.
        * The interval value represents the stimulus; if it is a list of values it is a list of stimuli.
        
        It should be pointed out that
        
        * Regardless of one or more stimulus it is here repesented in a list for the sake of not breaking previous code.
        * So this is a practical choice and does not contradict the above mentioned assumptions.
        
        """
        # Input population of size 100
        self.period = 5
        kHz_tone = lambda t0: list( np.linspace(t0, t0+self.period, 3) )
        self.t0 = 20
        origin = 0
        joinpulses = lambda headlist, taillist: [ headlist.append(i) for i in taillist ]
        joinInchnl = lambda origin, oldInchnl, pulse1: pulse1 if origin==0 else oldInchnl+pulse1
        self.input_channel = []
        self.stim_origin = [] # guide for plotting
        self.stim_end = []    # guide for plotting
        #all_intervals = [80, 130] #, 130, 180 #ms number of stimuli
        self.runtime = 0
        for an_interval in all_intervals:
            pulse1_t0 = origin + self.t0
            pulse2_t0 = pulse1_t0 + self.period + an_interval
            pulse1 = kHz_tone(pulse1_t0)
            pulse2 = kHz_tone(pulse2_t0)
            joinpulses(pulse1, pulse2)
            self.input_channel = joinInchnl(origin, self.input_channel, pulse1)
            self.stim_origin.append(origin)
            self.stim_end.append( pulse1[-1] )
            # update for next
            origin = pulse1[-1] + self.t0
            self.runtime = self.runtime + ( self.t0 + self.period + an_interval + self.period + self.t0 )

    # Generate input source
    def __gen_input_src(self):
        """The above created `self.input_channel` is used to create a spiking version of it.
        """
        spike_times = []
        for i in range( BuoMerz.n_input ):
            spike_times.append( self.input_channel )
        return sim.Population( BuoMerz.n_input, sim.SpikeSourceArray, {'spike_times': spike_times}, label="input" )

    # Private function for setting up the layers
    def __create_layers(self):
        """Creates layers: input, IV, III, and output.
        
        * The input layer is represented as a population of `self.n_input`.
        * Layer IV is represented as a dictionary with "exc" and "inh" for excitatory and inhibitory populations.
        * Similarly, for layer III.
        * The output layer is represented as a dictionary with keys of the form "out"+interval value, each representing a population.
        
        **Note:**
        
        * Interval value or the inter-pulse interval is the temporal information carried by the stimulus.
        * The interval value represents the stimulus; if it is a list of values it is a list of stimuli.
        
        """
        # input population
        popIn = sim.Population( BuoMerz.n_input, sim.IF_curr_alpha(), label="input" )
        
        # layer-4 receives input_src containing the PPF (paired-pulse facilitation)
        layer4 = { "exc": sim.Population( BuoMerz.n_ex4, sim.IF_curr_alpha(**BuoMerz.ex_cell_parameters), label="ex4" ),
                   "inh": sim.Population( BuoMerz.n_inh4, sim.IF_curr_alpha(**BuoMerz.inh_cell_parameters), label="inh4" ) }
        
        # layer-3 receives input from layer 4
        layer3 = { "exc": sim.Population( BuoMerz.n_ex3, sim.IF_curr_alpha(**BuoMerz.ex_cell_parameters), label="ex3" ),
                   "inh": sim.Population( BuoMerz.n_inh3, sim.IF_curr_alpha(**BuoMerz.inh_cell_parameters), label="inh3" ) }

        # output-layer receives input from excitatory population of layer 3
        output = {}
        #for i in np.nditer(BuoMerz.dual_pulse_intervals): # Buonomano and Merzenich tested the network for intervals between 30 and 330 ms with 10 ms steps
        for i in self.dual_pulse_intervals:
            unit_label = "output_for_"+str(i)
            unit_key = "out"+str(i)
            unit_val = sim.Population( BuoMerz.n_out, sim.IF_curr_alpha(**BuoMerz.ex_cell_parameters), label= unit_label )
            output.update( {unit_key: unit_val} )
        return [popIn, layer4, layer3, output]

    # Private function for setting up the wiring and connections
    def __connect(self):
        """ Based on Buonomano & Merzenich 1997 (Fig 3, p135)
        
        +----------------------+------------------+-----------------+------------------+-----------------+
        | Populations          | excitatory (III) | excitatory (IV) | inhibitory (III) | inhibitory (IV) |
        +======================+==================+=================+==================+=================+
        | **excitatory (III)** | 18/200 = 0.09    | -               | 12/200 = 0.06    | -               |
        +----------------------+------------------+-----------------+------------------+-----------------+
        | **excitatory (IV)**  | 18/120 = 0.15    | 15/120 = 0.125  | 12/120 = 0.1     | 10/120 = 0.083  |
        +----------------------+------------------+-----------------+------------------+-----------------+
        | **inhibitory (III)** | 8/50 = 0.16      | -               | 6/50 = 0.12      | -               |
        +----------------------+------------------+-----------------+------------------+-----------------+
        | **inhibitory (IV)**  | -                | 6/30 = 0.2      | -                | 4/30 = 0.133    |
        +----------------------+------------------+-----------------+------------------+-----------------+
        | **Input**            | -                | 15/100 = 0.15   | -                | 10/100 = 0.1    |
        +----------------------+------------------+-----------------+------------------+-----------------+
        
        **Note:**
        
        * The numerator correspond to the number of units (elements) within the recieving population.
        * The denominator is the convergence number of presynaptic inputs for each unit.
        * The decimal fraction are the probability values for the `sim.FixedProbabilityConnector` function.
        
        **Comments on connection with output layer:**
        
        The excitatory (III) population connects to the output layer `(1995) <https://doi.org/10.1126/science.7863330>`_ such that it proxies adaptation. The output layer populations recieves signal only at the end of the stimulus, i.e. during second pulse.
        
        Due to limitations with `SpNNaker8 <http://spinnakermanchester.github.io/>`_ here the connection is implemented such that it is made from the start of the simulation. However, only responses from the populations in the output layer at the end of the stimulus is returned.
        
        """
        wire_ex3ex3 = sim.FixedProbabilityConnector(0.09, allow_self_connections=True)
        wire_ex3inh3 = sim.FixedProbabilityConnector(0.06, allow_self_connections=False)
        wire_ex4ex3 = sim.FixedProbabilityConnector(0.15, allow_self_connections=False)
        wire_ex4ex4 = sim.FixedProbabilityConnector(0.125, allow_self_connections=True)
        wire_ex4inh3 = sim.FixedProbabilityConnector(0.1, allow_self_connections=False)
        wire_ex4inh4 = sim.FixedProbabilityConnector(0.083, allow_self_connections=False)
        wire_inh3ex3 = sim.FixedProbabilityConnector(0.16, allow_self_connections=False)
        wire_inh3inh3 = sim.FixedProbabilityConnector(0.12, allow_self_connections=True)
        wire_inh4ex4 = sim.FixedProbabilityConnector(0.2, allow_self_connections=False)
        wire_inh4inh4 = sim.FixedProbabilityConnector(0.133, allow_self_connections=True)
        wire_popInex4 = sim.FixedProbabilityConnector(0.15, allow_self_connections=False)
        wire_popIninh4 = sim.FixedProbabilityConnector(0.1, allow_self_connections=False)
        wire_inputpopIn = sim.FixedProbabilityConnector(0.25, allow_self_connections=False)
        # Below is custom made and not given by Buonomano & Merzenich
        wire_ex3output = sim.FixedProbabilityConnector(0.15, allow_self_connections=False)
        ### NOW CONNECT
        connect_ex3ex3 = sim.Projection( self.layer3["exc"], self.layer3["exc"], wire_ex3ex3,
                                         sim.StaticSynapse(weight=5.), receptor_type="excitatory" )
        connect_ex3inh3 = sim.Projection( self.layer3["exc"], self.layer3["inh"], wire_ex3inh3,
                                          sim.StaticSynapse(weight=5.), receptor_type="excitatory" )
        connect_ex4ex3 = sim.Projection( self.layer4["exc"], self.layer3["exc"], wire_ex4ex3,
                                         sim.StaticSynapse(weight=5.), receptor_type="excitatory" )
        connect_ex4ex4 = sim.Projection( self.layer4["exc"], self.layer4["exc"], wire_ex4ex4,
                                         sim.StaticSynapse(weight=2.5), receptor_type="excitatory" )
        connect_ex4inh3 = sim.Projection( self.layer4["exc"], self.layer3["inh"], wire_ex4inh3,
                                          sim.StaticSynapse(weight=5.), receptor_type="excitatory" )
        connect_ex4inh4 = sim.Projection( self.layer4["exc"], self.layer4["inh"], wire_ex4inh4,
                                          sim.StaticSynapse(weight=2.5), receptor_type="excitatory" )
        connect_inh3ex3 = sim.Projection( self.layer3["inh"], self.layer3["exc"], wire_inh3ex3,
                                          sim.StaticSynapse(weight=5.), receptor_type="inhibitory" )
        connect_inh3inh3 = sim.Projection( self.layer3["inh"], self.layer3["inh"], wire_inh3inh3,
                                           sim.StaticSynapse(weight=5.), receptor_type="inhibitory" )
        connect_inh4ex4 = sim.Projection( self.layer4["inh"], self.layer4["exc"], wire_inh4ex4,
                                          sim.StaticSynapse(weight=2.5), receptor_type="inhibitory" )
        connect_inh4inh4 = sim.Projection( self.layer4["inh"], self.layer4["inh"], wire_inh4inh4,
                                           sim.StaticSynapse(weight=2.5), receptor_type="inhibitory" )
        connect_popInex4 = sim.Projection( self.popIn, self.layer4["exc"], wire_popInex4,
                                           sim.StaticSynapse(weight=5.0), receptor_type="excitatory" )
        connect_popIninh4 = sim.Projection( self.popIn, self.layer4["inh"], wire_popIninh4,
                                            sim.StaticSynapse(weight=5.0), receptor_type="excitatory" )
        connect_inputpopIn = sim.Projection( self.input_src, self.popIn, wire_inputpopIn,
                                             sim.StaticSynapse(weight=1.0), receptor_type="excitatory" )
        # connect the exc3 unit with output layer
        #for i in np.nditer(BuoMerz.dual_pulse_intervals): # Buonomano and Merzenich tested the network for intervals between 30 and 330 ms with 10 ms steps
        for i in self.dual_pulse_intervals:
            if i == self.inter_pulse_interval:
                unit_key = "out"+str(i)
                prj = [ sim.Projection( self.layer3["exc"], self.output[unit_key], wire_ex3output,
                                        sim.StaticSynapse(weight=100.0), receptor_type="excitatory" )
                       for k in self.output.keys() if k==unit_key ]
                        
    # Private function for recording
    def __record(self):
        """Records into `Neo <https://neo.readthedocs.io/en/latest/>`_ object for the four layers: input, IV, III, and output.
        """
        self.input_src.record("all")
        self.popIn.record("all")
        rec = lambda pop: [ subpop.record("all") for subpop in pop.values() ]
        rec( self.layer4 )
        rec( self.layer3 )
        rec( self.output )

    # Private function for extracting recorded data
    def __getdata(self):
        """
        Gets the recorded `Neo <https://neo.readthedocs.io/en/latest/>`_ objects.
        The only exception is for the output layer. Because of the reasons given above, for the output layer rather than a Neo object its `SpikeTrains` are returned. They are the responses at the end of the stimulus. To help visualizing them along with `SpikeTrains` from other layers a padding is also returned.
        """
        #spikes_input_src = self.input_src.get_data("spikes")
        neo_popIn = self.popIn.get_data(variables=["spikes", "v"])
        neo_ex4 = self.layer4["exc"].get_data(variables=["spikes", "v"])
        neo_inh4 = self.layer4["inh"].get_data(variables=["spikes", "v"])
        neo_ex3 = self.layer3["exc"].get_data(variables=["spikes", "v"])
        neo_inh3 = self.layer3["inh"].get_data(variables=["spikes", "v"])
        # For the proxy output layer get just the spikes corresponding to the second pulse
        neo_out = {}
        for key in self.output.keys():
            #neo_out.update( { key: self.output[key].get_data( variables=["spikes", "v"] ) } )
            for_all_neurons = self.output[key].get_data( variables=["spikes"] )
            spktrains_all = []
            for spktrain in for_all_neurons.segments[0].spiketrains:
                spktrains_all.append( spktrain[ spktrain > self.stim_end[0]*pq.ms] ) # spikes for 2nd pulse only
            neo_out.update( { key: {"spiketrains": spktrains_all,
                                    "placeholder_axes": self.__placeholder_x_y(neo_ex3) } } )
        return [neo_popIn, neo_ex4, neo_inh4, neo_ex3, neo_inh3, neo_out]

    # Private function for dummy time and y-axis for output layer
    def __placeholder_x_y(self, neo_ex3):
        """
        Returns a list containing two lists. The first represent t-axis from zero to start of second pulse of the stimuli. The second is an array of zeros whose size is the same as the first.
        """
        spks = neo_ex3.segments[0].spiketrains[0] # other neo_xyz objects should also work
        # tstart = spks.t_start or tstart = spks.t_start + self.t0 + self.period
        tstart = spks.t_start #+ self.t0*pq.ms + self.period*pq.ms
        ndata = spks.sampling_rate * ( spks.t_stop - tstart )
        taxis = np.linspace( tstart, spks.t_stop, num=ndata.magnitude )
        return [ taxis, np.zeros( taxis.shape ) ]
    
    # Private function for checking and getting the interval for plotting all the layers
    def __get_interval_of_interest(self, variable_argument_intv_tuple):
        """."""
        if len(variable_argument_intv_tuple)==0 and len(self.str_dual_pulse_intervals)==1:
            return self.str_dual_pulse_intervals()[0]
        elif (len(variable_argument_intv_tuple)==0 and len(self.str_dual_pulse_intervals)>1) or (len(variable_argument_intv_tuple)>1):
            raise ValueError("Argument must be a string representing single interval (stimulus).")
        else:
            if type( variable_argument_intv_tuple[0] ) is str:
                return variable_argument_intv_tuple[0]
            else:
                return str( variable_argument_intv_tuple[0] )
        
    # Plotting functions
    def plot_all_layers(self, *intv):
        """Visualize spikes for populations in all layers; starting from below: input, excitatory in LayerIV, inhibitory in LayerIV, excitatory in LayerIII, inhibitory in LayerIII and lastly **a** population in the output layer at the top.
        y-axis are labelled with respective population and x-axis is the time in milliseconds.
        """
        y = self.data_for_all_intervals
        #z = "80"
        z = self.__get_interval_of_interest(intv)
        #
        fig, ( (sp1),
               (sp2),
               (sp3),
               (sp4),
               (sp5),
               (sp6) ) = plt.subplots(6,1,sharex=True)
        fig.suptitle("Temporal Info. Processing for S = "+z+" ms")

        [ sp1.plot(y[z]["out"]["out"+z]["placeholder_axes"], alpha=0.0) if i==0 else
          sp1.eventplot( y[z]["out"]["out"+z]["spiketrains"] ) for i in range(2) ]
        sp1.set(ylabel="output")

        sp2.eventplot( y[z]["inh3"].segments[0].spiketrains )
        sp2.set(ylabel="inh3")

        sp3.eventplot( y[z]["ex3"].segments[0].spiketrains )
        sp3.set(ylabel="ex3")

        sp4.eventplot( y[z]["inh4"].segments[0].spiketrains )
        sp4.set(ylabel="inh4")

        sp5.eventplot( y[z]["ex4"].segments[0].spiketrains )
        sp5.set(ylabel="ex4")

        sp6.eventplot( y[z]["popIn"].segments[0].spiketrains )
        sp6.set(ylabel="popIn")
        sp6.set(xlabel="time (ms)")

        plt.show()
    
    def plot_output_layer_timeseries(self):
        """Visualize spikes from each population in the output layer as a time-series.
        y-axis represent each unit in a population and x-axis is time in milliseconds.
        """
        y = self.data_for_all_intervals
        intvs = self.str_dual_pulse_intervals #["80", "130", "180", "230", "280"]
        clrs = ['C{}'.format(i) for i in range(len(intvs))] # colors for respective interval
        legpatches = []
        fig, ( (sp) ) = plt.subplots(1,1,sharex=True)
        fig.suptitle("Populations (output layer) triggered by respective stimulus")
        for j in range( len(intvs) ):
            [ sp.plot(y[intvs[j]]["out"]["out"+intvs[j]]["placeholder_axes"], alpha=0.0) if i==0 else
              sp.eventplot( y[intvs[j]]["out"]["out"+intvs[j]]["spiketrains"], colors=clrs[j] ) for i in range(2) ]
            legpatches.append( mpatches.Patch(color=clrs[j], label=intvs[j]) )
        sp.set(ylabel="unit in population")
        sp.set(xlabel="time (ms)")
        sp.legend( handles=legpatches, shadow=True )
    
    def plot_output_layer_vertical(self):
        """Visualize spikes from each population in the output layer as one-above-eachother.
        Left y-axis represent the population and right y-axis the stimulus.
        """
        y = self.data_for_all_intervals
        intvs = self.str_dual_pulse_intervals #["80", "130", "180", "230", "280"]
        #
        mrows = len(intvs) # mrows of subplots
        clrs = ['C{}'.format(i) for i in range(mrows)] # mrows of colors
        poplabels = self.__population_labels(mrows)
        #
        fig, splist = plt.subplots( mrows, 1 )
        fig.suptitle("Populations (output layer) triggered by respective stimulus")
        for i in range(mrows):
            z = intvs[i]
            # plot
            splist[i].eventplot( y[z]["out"]["out"+z]["spiketrains"],
                                 colors=clrs[i] )
            splist[i].margins(10, 0.5) # zoom-out
            # legend
            z_patch = mpatches.Patch( color=clrs[i], label=z )
            splist[i].legend( handles=[z_patch], shadow=True )
            # left yticks and ylabel
            splist[i].set_yticks( [] )
            splist[i].set( ylabel="output\npop-"+poplabels[i] )
            # right yticks and ylabel
            rightside = splist[i].twinx()
            rightside.set_yticks( [] )
            rightside.set_ylabel( "stimulus\n"+z, multialignment='center' )
        # remove vertical gap between subplots
        plt.subplots_adjust( hspace=.0 )
        plt.xticks( [] ) # remove xticks
        plt.show()
    
    # Private function for generating labels for neuron populations
    def __population_labels(self, mrows):
        """Generates ``mrows`` of English letters from "a" to ...
        This is the naive method shown in https://www.geeksforgeeks.org/python-ways-to-initialize-list-with-alphabets/
        """
        labellist = []
        alpha = "a"
        for i in range(0, mrows):
            labellist.append( alpha )
            alpha = chr( ord(alpha) + 1 )
        return labellist
    
    def plot_output_layer_timeseries_vertical(self):
        """Visualize spikes from each population in the output layer as a time-series.
        But unlike :py:meth:`plot_output_layer_timeseries` this one plots one-above-eachother.
        Left y-axis represent units in a population and right y-axis the respective population.
        """
        y = self.data_for_all_intervals
        intvs = self.str_dual_pulse_intervals #["80", "130", "180", "230", "280"]
        #
        mrows = len(intvs) # mrows of subplots
        clrs = ['C{}'.format(i) for i in range(mrows)] # mrows of colors
        poplabels = self.__population_labels(mrows)
        #
        fig, splist = plt.subplots( mrows, 1, sharex=True )
        fig.suptitle("Populations (output layer) triggered by respective stimulus")
        for j in range(mrows):
            z = intvs[j]   # a stimulus
            sp = splist[j] # a subplot
            # plot
            [ sp.plot(y[z]["out"]["out"+z]["placeholder_axes"], alpha=0.0) if i==0 else
              sp.eventplot( y[z]["out"]["out"+z]["spiketrains"], colors=clrs[j] ) for i in range(2) ]
            # legend
            z_patch = mpatches.Patch(color=clrs[j], label=z)
            sp.legend( handles=[z_patch], shadow=True )
            # xticks and xlabel only for bottom subplot
            if j==(mrows-1):
                sp.set(xlabel="time (ms)")
            #else:
            #    sp.set_xticks( [] )
            # left ylabel
            sp.set(ylabel='units in\npop')
            # right yticks and ylabel
            rightside = sp.twinx()
            rightside.set_yticks( [] )
            rightside.set_ylabel( "pop-"+poplabels[j] )
        plt.subplots_adjust( hspace=.0 )
        plt.show()